package control;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.io.File;
import java.util.ArrayList;
import java.util.StringTokenizer;

import gui.panel.ProcessModel;
import org.jdom.input.SAXBuilder;
import org.jdom.Document;
import org.jdom.JDOMException;
import org.jdom.Element;

/**
 * The configloader class reads the configfile when running the testbench without gui.
 * This class is responsible for reading the configfile, loading the specified operations
 * and setting correct properties for each operation.
 *
 * @author Ole Kristian Fivelstad
 */
public class ConfigLoader {

    private Control control;
    private ProcessModel model;

    /**
     * Constructor for the configloader class.
     *
     * @param control instance of the control class
     */
    public ConfigLoader(Control control){
        model = new ProcessModel();
        this.control=control;
    }

    /**
     * Method for getting the chosen process. This process is given as an
     * arraylist of operations.
     *
     * @return ArrayList containing the process
     */
    public ArrayList getOperations(){
        return model.getProcess();
    }

    /**
     * Method for reading the configfile.
     * Adds chosen operations to the processmodel and sets properties.
     */
    public void readConfigFile(){

        ArrayList fitiProps = new ArrayList();
        ArrayList filterProps = new ArrayList();
        ArrayList collocationProps = new ArrayList();
        ArrayList wsdProps = new ArrayList();
        ArrayList conceptProps = new ArrayList();
        SAXBuilder sb = new SAXBuilder();
        Document doc=null;
        try {
            doc = sb.build("data"+File.separator+"testbench_config.xml");

            Element root = doc.getRootElement();

            Element dataSet = root.getChild("dataset");
            String dataSetURL = dataSet.getValue();
            control.setDataset(dataSetURL.trim());

            Element operations = root.getChild("operations");
            String operationString = operations.getValue();
            StringTokenizer st = new StringTokenizer(operationString);
            while(st.hasMoreTokens()){
                int operation = Integer.valueOf(st.nextToken());
                model.addOperation(operation-1);
            }

            Element parameters = root.getChild("parameters");

            //WSD-properties
            Element adaptedLesk = parameters.getChild("adaptedlesk");
            wsdProps.add(new OperationProperty("Use adapted Lesk",adaptedLesk.getValue()));
            Element contextSize = parameters.getChild("contextsize");
            wsdProps.add(new OperationProperty("Size of context",contextSize.getValue()));
            Element keepVerbs = parameters.getChild("keepverbs");
            wsdProps.add(new OperationProperty("Keep verbs", keepVerbs.getValue()));
            Element keepAdj = parameters.getChild("keepadjectives");
            wsdProps.add(new OperationProperty("Keep adjectives", keepAdj.getValue()));

            //Concept-properties
            Element parentnode = parameters.getChild("parentnode");
            conceptProps.add(new OperationProperty("Add parent node",parentnode.getValue()));
            Element maxDistance = parameters.getChild("maxdistance");
            conceptProps.add(new OperationProperty("Max distance in hierarchy",maxDistance.getValue()));
            Element keepOriginal = parameters.getChild("keeporiginalterms");
            conceptProps.add(new OperationProperty("Keep original terms", keepOriginal.getValue()));

            //Collocation-properties
            Element verbs = parameters.getChild("verbs");
            collocationProps.add(new OperationProperty("Keep verbs", verbs.getValue()));
            Element adv = parameters.getChild("adverbs");
            collocationProps.add(new OperationProperty("Keep adverbs", adv.getValue()));
            Element adj = parameters.getChild("adjectives");
            collocationProps.add(new OperationProperty("Keep adjectives", adj.getValue()));
            Element nums = parameters.getChild("numbers");
            collocationProps.add(new OperationProperty("Keep numbers", nums.getValue()));


            //Fiti-properties
            Element minsupport = parameters.getChild("minsupport");
            fitiProps.add(new OperationProperty("Minumum support",minsupport.getValue()));
            Element maxsupport = parameters.getChild("maxsupport");
            fitiProps.add(new OperationProperty("Maximum support",maxsupport.getValue()));
            Element minconfidence = parameters.getChild("minconfidence");
            fitiProps.add(new OperationProperty("Minumum confidence",minconfidence.getValue()));
            Element maxconfidence = parameters.getChild("maxconfidence");
            fitiProps.add(new OperationProperty("Maximum confidence",maxconfidence.getValue()));
            Element maxsetsize = parameters.getChild("setsize");
            fitiProps.add(new OperationProperty("Maximum set size", maxsetsize.getValue()));
            Element maxspan = parameters.getChild("maxspan");
            fitiProps.add(new OperationProperty("Maxspan", maxspan.getValue()));

            //Filter-properties
            Element terms = parameters.getChild("filterterms");
            filterProps.add(new OperationProperty("Maximum number of terms", terms.getValue()));
            Element weight = parameters.getChild("getbyweight");
            filterProps.add(new OperationProperty("Get by weight", weight.getValue()));



        }
        catch (Exception e) {
            e.printStackTrace();
            System.out.println("Something went wrong when building dataset...");
        }

        //Setting properties for the operations
        for(int i=0;i<model.getProcess().size();i++){
            Operation op = model.getOperation(i);
            if(op.getClass().getName().indexOf("Fiti")!=-1){
                System.out.println("Setting FITI-properties");
                op.setProperties(fitiProps);
            } else if(op.getClass().getName().indexOf("Filter")!=-1){
                System.out.println("Setting Filter-properties");
                op.setProperties(filterProps);
            } else if(op.getClass().getName().indexOf("Collocation")!=-1){
                System.out.println("Setting Collocation-properties");
                op.setProperties(collocationProps);
            } else if(op.getClass().getName().indexOf("Concept")!=-1){
                System.out.println("Setting Concept-properties");
                op.setProperties(conceptProps);
            } else if(op.getClass().getName().indexOf("WordSenseDisambiguation")!=-1){
                System.out.println("Setting WSD-properties");
                op.setProperties(wsdProps);
            }
        }
    }
    /**
     * Internal class defining a structure used to store possible operations.
     *
     * @author Kjell-Inge Skogstad and Trond ivind Eriksen
     */
    private class Item {

        private String description;
        private Class c;

        /**
         * Constructor storing operation description and class.
         *
         * @param description The description
         * @param c The class
         */
        public Item(String description, Class c) {
            this.description = description;
            this.c = c;
        }
    }


}
