package control;

import java.util.ArrayList;
import java.io.File;
import java.io.FileInputStream;

import modelling.ModelOperation;
import dataloader.FileHandler;
import datamodel.*;
import datapreparation.TextOperation;
import gui.MainFrame;
import gui.MainFrameDummy;
import newsloader.NewsLoader;
import net.didion.jwnl.JWNL;


/**
 * The control class handling the application logic. This class is responsible
 * for handling the temporal text mining process and initializes the rest of the 
 * application.
 * 
 * @author Kjell-Inge Skogstad and Trond ivind Eriksen
 */
public class Control extends Thread {

    private DataSet dataSet;
    private MainFrame mainFrame;
    private ArrayList docReports;
    private ConfigLoader loader;
    private String dataSetUrl;


    /**
     * Constructor that initializes the gui.
     */
    public Control(boolean gui) {
        docReports = new ArrayList();
        Globals.showGui=gui;
        if(!gui){
            System.out.println("Running control...");
            loader = new ConfigLoader(this);
            loader.readConfigFile();
            MainFrameDummy mainFrameDummy = new MainFrameDummy(this);
            System.out.println("Loading dataset...");
            openDataSet(dataSetUrl);
            System.out.println("Dataset loaded");
            System.out.println("Running process...");
            runProcess();
        }else{
            /* Initializes the gui */
            mainFrame = new MainFrame(this);
        }
    }

    public void setDataset(String url){
        dataSetUrl="data"+ File.separator+url;
    }

    /**
     * Method for setting the data set to be used in the application. This 
     * method uses a reference to the source of the dataset (file) and use
     * a file handler to retrieve the dataset.
     * 
     * @param location The location of the dataset file
     */
    public void openDataSet(String location) {
        try {
            dataSet = FileHandler.loadDataSet(location);

            /* Update gui */
            generateDocReports();


            /* Updating name, language and nofTexts */
            if(Globals.showGui){
            mainFrame.configGui(
                    dataSet.getTexts().size(),
                    dataSet.getName(),
                    dataSet.getLanguageStr(),
                    dataSet.getNofTexts() + "");
            }
        } catch (Exception e) {
            if(Globals.showGui){
                mainFrame.printMessage("DataSet not available.");
            }
            e.printStackTrace();
        }
    }

    /**
     * Method for saving a data set. The location sets the file to which
     * the data set shall be saved.
     * 
     * @param location The location
     */
    public void saveDataSet(String location) {
        FileHandler.storeDataSet(location, dataSet);
    }

    /**
     * Method for saving a report. The location sets the file to which the 
     * report shall be saved.
     * 
     * @param location The location
     * @param report The report
     */
    public void saveReport(String location, String report) {
        String addedDataSetInfo = "";
        String top = "";

        top += "\n********** TTM Testbench Results *********************************************\n\n";

        addedDataSetInfo += "Dataset: \n";
        addedDataSetInfo += "Name: " + dataSet.getName() + "\n";
        addedDataSetInfo += "Time Granularity: " + dataSet.getTimeGranulity() + "\n";
        addedDataSetInfo += "Language: " + dataSet.getLanguageStr()+ "\n";
        addedDataSetInfo += "DataSet size: " + dataSet.getNofTexts() + "\n";
        addedDataSetInfo += "\n";        

        report = top + addedDataSetInfo + report;

        FileHandler.storeReport(location, report);
    }

    /**
     * Method for creating a new set of reports, one for each document, to be
     * shown in gui.
     */
    private void generateDocReports() {
        Text text;

        docReports = new ArrayList();
        if (dataSet != null) {
            for (int i = 0; i < dataSet.getTexts().size(); i++) {
                String report = "";

                text = (Text)dataSet.getTexts().get(i);

                report += "DataSet: " + dataSet.getName() + "\n";
                report += "Document time: " + text.getTid() + " (" +
                            dataSet.getTimeGranulity() + ")\n\n";
                report += "Text: \n" + text.toString() + "\n";

                /* 
                * If terms have been extracted previously print these as
                * well
                */
                if (text.getTerms().size() > 0) {
                    report += "\n";
                    report += "Terms: (" + text.getTerms().size() + ")\n" +
                                text.getPrintableTerms() + "\n";
                }

                docReports.add(report);
            }
        }
    }

    /**
     * Returns the report of a specific document given by its index.
     * 
     * @param index The index
     * @return The report
     */
    public String getDocReport(int index) {
        String report = "";

        /* Test if index is legal */
        if (docReports.size() > index) {
            report = (String)docReports.get(index);
        }

        return report;
    }

    /**
     * Method used when a data preparation operation has been performed. Updates
     * the report with the new list of terms.
     * 
     * @param index The document index
     */
    private void updateDocReport(int index) {
        Text text;

        /* Tests if a dataset is present */
        if (dataSet != null) {
            for (int i = 0; i < dataSet.getTexts().size(); i++) {
                String report = (String)docReports.remove(i);

                text = (Text)dataSet.getTexts().get(i);

                report += "\n";
                report += "Operation: " + index + "\n";
                report += "Terms: (" + text.getTerms().size() + ")\n" +
                            text.getPrintableTerms() + "\n";

                docReports.add(i, report);
            }
        }
    }

    /**
     * Method used to start the temporal text mining process in a separate
     * thread.
     */
    public void runProcess() {
        new Thread(this).start();
    }

    /**
     * Run method containing the process logic. This method is to be started by
     * the runProcess() method.
     */
    public void run() {
        if (dataSet != null) {
            ArrayList operations;
            if(!Globals.showGui){
                operations = loader.getOperations();
            }
            else{
                operations = mainFrame.getProcess();
            }
            /* Tests if a valid list of operations is present */
            if (operations.size() > 0) {
                Operation op;

                try {
                    String propsFile = "jwnl"+File.separator+"file_properties.xml";
                    JWNL.initialize(new FileInputStream(propsFile));
                } catch (Exception e) {
                    System.out.println("JWNL data not available...");
                    e.printStackTrace();
                    System.exit(0);
                }

                /* Perform operation and update gui */
                for (int i = 0; i < operations.size(); i++) {
                    op = (Operation)operations.get(i);

                    /* Perform operation */
                    if(!Globals.showGui){
                        System.out.println("Performing operation: "+op.getClass().getSimpleName()+"...");
                    }
                    //long startTime = System.currentTimeMillis();
                    op.performOperation(dataSet);
                    //long endTime = System.currentTimeMillis();
                    //double timeSpent = ((double)(endTime-startTime))/1000;
                    //System.out.println(op.getClass().getSimpleName()+" used "+timeSpent+" seconds.");

                    /* Test on type of operation */
                    if (op instanceof TextOperation) {
                        updateDocReport(i + 1);
                    } else if (op instanceof ModelOperation) {
                        ((ModelOperation)op).printResults();
                    }

                    /* Log result */
                    if (op.getLogResult()) {
                        //not implemented
                    }

                    /* Update progressbar */
                    if(Globals.showGui){
                        mainFrame.updateOperationProgress(0);
                        mainFrame.updateProgress((int)((double)(i + 1) /
                            (double)operations.size() * 100));
                    }
                }
                /* Update gui */
                if(Globals.showGui){
                    mainFrame.setSliderIndex(1);
                }
            } else {
                if(Globals.showGui){
                    mainFrame.printMessage("Process operations not defined.");
                }
            }
        } else {
            if(Globals.showGui){
                mainFrame.printMessage("DataSet not available.");
            }
            else{
                System.out.println("DataSet not available");
            }
        }
    }

    /**
     * Main method initializing a new control instance.
     * 
     * @param args Arguments
     */
    public static void main(String[] args) {
        if (args.length == 1) {
            if (args[0].equalsIgnoreCase("docloader")) {
                new NewsLoader();
            } else {
                System.out.println("Please see user manual for correct arguments.");
            }
        } else if (args.length == 2) {
            if (args[0].equalsIgnoreCase("testbench")) {
                if(args[1].equalsIgnoreCase("false")){
                    System.out.println("No GUI");
                    new Control(false);
                }
                else if(args[1].equalsIgnoreCase("true")){
                    new Control(true);
                }
                else{
                    System.out.println("Please see user manual for correct arguments.");
                }

            }
            else{
                System.out.println("Please see user manual for correct arguments.");
            }
        } else {
            System.out.println("Please see user manual for correct arguments.");
        }
    }
}
