package control;

import java.util.ArrayList;

import datapreparation.partofspeech.PartOfSpeech;
import datapreparation.filter.Filter;
import datapreparation.noun.Noun;
import datapreparation.stemming.Stemming;
import datapreparation.stopword.Stopword;
import datapreparation.tokinize.Tokinize;
import datapreparation.weighting.Weighting;
import datapreparation.wordsensedisambiguation.WordSenseDisambiguation;
import datapreparation.concept.Concept;
import datapreparation.collocation.Collocation;
import modelling.fiti.Fiti;

/**
 * Class used to configurate the possible operations to be used in the 
 * application. New operations added should be referenced and added to the 
 * list in this class.
 * 
 * @author Kjell-Inge Skogstad and Trond ivind Eriksen
 */
public class OperationConfig {

    private ArrayList operations;

    /**
     * Constructor creating a list of possible operations. 
     */
    public OperationConfig() {
        operations = new ArrayList();

        /* Add possible operations here */        
        addOperation(new Item("Tag texts with POS", PartOfSpeech.class));
        addOperation(new Item("Extract collocations", Collocation.class));
        addOperation(new Item("Disambiguate nouns", WordSenseDisambiguation.class));
        addOperation(new Item("Extract concepts", Concept.class));
        addOperation(new Item("Weight words", Weighting.class));
        addOperation(new Item("Filter terms", Filter.class));
        addOperation(new Item("FITI", Fiti.class));

    }
        

    /**
     * Method to be used for adding operation to the list.
     * 
     * @param operation The operation
     */
    public void addOperation(Item operation) {
        operations.add(operation);
    }

    /**
     * Method used by the application to get a list of the operations available.
     * This list only contains the textual description of the operation.
     * 
     * @return Descriptions of the operations
     */
    public ArrayList getOperationDescriptions() {
        ArrayList descriptions = new ArrayList();

        for (int i = 0; i < operations.size(); i++) {
            descriptions.add(((Item) operations.get(i)).description);
        }
        return descriptions;
    }

    /**
     * Method used by the application to retrieve the class which implements 
     * the operation.
     * 
     * @param description The operation description
     * @return The operation class
     */
    public Class getOperation(String description) {
        Class c = null;

        for (int i = 0; i < operations.size(); i++) {
            if (description.equals(((Item) operations.get(i)).description)) {
                c = ((Item) operations.get(i)).c;
                break;
            }
        }
        return c;
    }

    /**
     * Method used by the application to retrieve the description of a operation
     * given by its implementation class.
     * 
     * @param c The class
     * @return The description
     */
    public String getOperationDescription(Class c) {
        String desc = "";
        Item item;

        for (int i = 0; i < operations.size(); i++) {
            item = (Item) operations.get(i);

            if (c.equals(item.c)) {
                return item.description;
            }
        }

        return desc;
    }

    /**
     * Method for testing if two operations are equal (of the same type).
     * 
     * @param op1 Operation 1
     * @param op2 Operation 2
     * @return True if equal
     */
    public boolean isEqual(Operation op1, Operation op2) {
        if ((op1 instanceof Filter && op2 instanceof Filter) ||
                (op1 instanceof Fiti && op2 instanceof Fiti) ||
                (op1 instanceof Noun && op2 instanceof Noun) ||
                (op1 instanceof Stemming && op2 instanceof Stemming) ||
                (op1 instanceof Stopword && op2 instanceof Stopword) ||
                (op1 instanceof Tokinize && op2 instanceof Tokinize) ||
                (op1 instanceof Weighting && op2 instanceof Weighting)) {
            return true;
        }
        return false;
    }

    /**
     * Internal class defining a structure used to store possible operations. 
     * 
     * @author Kjell-Inge Skogstad and Trond ivind Eriksen
     */
    private class Item {

        private String description;
        private Class c;

        /**
         * Constructor storing operation description and class.
         * 
         * @param description The description
         * @param c The class
         */
        public Item(String description, Class c) {
            this.description = description;
            this.c = c;
        }
    }
}
