package dataloader;

import java.util.ArrayList;
import java.lang.Integer;

/**
 * This is a straightforward implementation of the famous Knuth-Morris-Pratt 
 * algorithm for string-matching. We use this to algorithm to find the  
 * indices where each news in the HTML files starts.
 * 
 * @author Trond ivind Eriksen and Kjell-Inge Skogstad
 */
public class KMP{
	
	/**
	 * KMP is not a particularily fast algorithm in practice, but it achieves 
	 * an linear upper bound complexity. The algorithm basically constructs a 
	 * prefixarray for the pattern, and uses this to test the different 
	 * possibilities of a match in linear time. It is not obvious that the 
	 * complexity is linear because we see two nested loops in the algorithm. 
	 * But once we notice that q can only be incremented by 1 each pass of 
	 * the mainloop. Since the inner loop can only be run until q is zero and 
	 * the prefix is a strictly increasing function this implies that the overall 
	 * complexity is linear according to basic amortized complexity. 
	 * See Cormen et al: Introduction to algorithms for a formal proof.
	 * @param text
	 * @param pattern
	 * @return ret
	 */
	public int[] kmp(String text, String pattern){
		ArrayList<Integer> hits = new ArrayList<Integer>();
		int[] prefix = computePrefix(pattern);
		int n = text.length();
		int m = pattern.length();
		int q=0;
		for(int i=0;i<n;i++){
			while(q>0 && pattern.charAt(q)!= text.charAt(i)){
				q=prefix[q-1];
			}
			if(pattern.charAt(q)==text.charAt(i)) q++;
			if(q==m){
				hits.add(i-m+1);
				q=prefix[q-1];
			}
		}
		int[] ret = new int[hits.size()];
		for(int i=0;i<hits.size();i++) ret[i]=hits.get(i);
		return ret;
	}
	
	/**
	 * This function computes the prefixfunction of the pattern. 
	 * It is basically a kind of index in the pattern. The complexity is linear here as well.
	 * There is a similar proof of the complexity here as in the main algorithm.
	 */
	private int[] computePrefix(String pattern){
		int m = pattern.length();
		int[] prefix = new int[m];
		prefix[0]=0;
		int k=0;
		for(int q=1;q<m;q++){
			while(k>0 && pattern.charAt(k)!=pattern.charAt(q)) k=prefix[k-1];
			if(pattern.charAt(k)==pattern.charAt(q))k++;
			prefix[q]=k;
		}
		return prefix;
	}
	
}//class