package dataloader;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.StringTokenizer;

import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;

import datapreparation.language.LanguageDecider;

/**
 * The class reads config.xml and gets the URL or catalogue with the HTML-files
 * one wishes to tokenize. These are then sent to HTMLStripper.
 * When HTMLStripper has tokenized the documents, XMLParser generates the file
 * tokenized.xml
 * 
 * @author Kjell-Inge Skogstad and Trond ivind Eriksen
 */
public class XMLParseren {
	
	private boolean isFileMade = false;
	private File oldFile = new File("tokenized.xml");
	HTMLStripper hs ;
	private boolean fromUrl;
	private int counter_docs;
	private int timestamp_docs = 0;
	private int numberOfTexts = 0;
	private LanguageDecider ld;
	
	
	/**
	 * The constructor to the class. Reads the XML file that is to be parsed.
	 * 
	 * @param file File the config file
	 * @param fromUrl Checks if the input files comes from file or from url.
	 */
	public XMLParseren(File file, boolean fromUrl){	
		this.fromUrl = fromUrl;
		readXML(file);
	}
	
    /**
     * Constructor, reads the XML file that is to be parsed.
     * 
     * @param fromUrl The config file
     * @param toUrl The dataset file
     */
    public XMLParseren(String fromUrl, String toUrl) {
        this.fromUrl = true;
        oldFile = new File(toUrl);
        readXML(new File(fromUrl));
    }
    
	/**
	 * The method reads the content in the configuration file (the XML file)
	 * @param file The input file
	 * @return The doclist with the documents
	 */
	public ArrayList readXML(File file){
		SAXBuilder myBuilder = new SAXBuilder();
		Document myDocument = null;
		try {
			myDocument = myBuilder.build(file);
		}
		
		catch (JDOMException jde) {
			jde.printStackTrace();
			
		} catch (IOException e) {
			e.printStackTrace();
		}
		
		/* Retrieves information from the XML file */
		Element root = myDocument.getRootElement();
		Element collection = root.getChild("collection");
		Element uri = collection.getChild("uri");
		Element directory = collection.getChild("directory");
		Element tokenizer = root.getChild("tokenizer");
		Element sentenceboundary = tokenizer.getChild("sentenceboundary");
		Element paragraphboundary = tokenizer.getChild("paragraphboundary");
		Element allsmall = tokenizer.getChild("allsmall");
		
		String urls = uri.getText();
		String directoryText = directory.getText();
		String sBoundary = sentenceboundary.getText();
		String pBoundary = paragraphboundary.getText();
		String aSmall = allsmall.getText();
		
		if(fromUrl){
			ArrayList uriList = parseUri(urls);
			ArrayList docList = new ArrayList();
			for(int i = 0; i < uriList.size(); i++){
				hs = new HTMLStripper((String)uriList.get(i), sBoundary, pBoundary, aSmall, this);
				String orginalfilen = hs.getOriginalFile();
				String parsetFil = hs.parseFilen(orginalfilen);	
			}
			return docList;
		}
		
		else{
			hs = new HTMLStripper(directoryText, this, sBoundary, pBoundary ); // Denne kjøres dersom det skal hentes fra fil
			System.err.println("readXML i XMLParser");
			return null;
		}
	}
	
	/**
	 * The method finds all URLs that were listed in the configuration file.
	 * @param urls The string of URLs.
	 * @return The list of URIs
	 */
	public ArrayList parseUri(String urls){
		StringTokenizer st = new StringTokenizer(urls);
		ArrayList uriList = new ArrayList();
		while(st.hasMoreTokens()){
			uriList.add(st.nextToken());
		}
		return uriList;
	}
	
	/**
	 * The method creates the tokenized.xml file.
	 * @param title
	 * @param url
	 * @param body
	 * @throws FileNotFoundException
	 * @throws SecurityException
	 */
	public void makeXmlTokenizer(String title, String url, String body) throws FileNotFoundException, SecurityException{
		String fileContent = "";
		try {
			fileContent = readFile(oldFile);
		} catch (IOException e) {
			//e.printStackTrace();
            //Not dangerous
		}
		String encoding = "ISO-8859-1";
		fileContent = fileContent.replaceAll("<\\?xml version=\"1.0\" encoding=\""+ encoding+"\"\\?>", "");
		fileContent = fileContent.replaceAll("<tokenized>", "");
		fileContent = fileContent.replaceAll("</tokenized>", "");
		
		PrintWriter out = new PrintWriter(new FileOutputStream("tokenized.xml"));
		
		out.println("<?xml version=\"1.0\" encoding=\""+ encoding+"\"?>");
		out.println("<tokenized>");
		out.println(fileContent);
		out.println("<doc id = \"" + url + "\">");
		out.println("<title>");
		out.println(title);
		out.println("</title>");
		out.println("<body>");
		out.println(body);
		out.println("</body>");
		out.println("</doc>");
		out.println("</tokenized>");
		out.flush();
		out.close();
	}
	
	/**
	 * The method creates tokenized.xml, the file of news to be loaded into 
	 * the system
	 * @param text The text
	 * @param url The URL
	 * @throws FileNotFoundException
	 * @throws SecurityException
	 */
	public void makeXmlTokenizer2(String text, String url) throws FileNotFoundException, SecurityException{
		String fileContent = "";
		timestamp_docs +=1;
		
		try {
			fileContent = readFile(oldFile);
		} catch (IOException e) {
			//e.printStackTrace();
            //Not dangerous
		}
		
		PrintWriter out = new PrintWriter(new FileOutputStream(oldFile));
		
		String name = "FTnews";					//the name
		String dataSetTime = "days";			//days, hours, weeks
		String language ="";					//language of the documents
		
		/** Building the xml file*/
		if(isFileMade==false){
			/* Finds the language to the document collection */
			boolean languageDecided = false;
			if(!languageDecided){
				String firstNews = (String) hs.getNewsList().get(0);
				ld = new LanguageDecider();
				int theLanguage = ld.checkSentenceInLanguage(firstNews);
				if(theLanguage == 0) language = "English";
				if(theLanguage == 1) language = "Norwegian";
				if(theLanguage == 2) language = "Undecided";
			}
			
			/* Creating the XML file */
			String encoding = "ISO-8859-1";
			fileContent = fileContent.replaceAll("<\\?xml version=\"1.0\" encoding=\""+ encoding+"\"\\?>", "");
			fileContent = fileContent.replaceAll("<tokenized>", "");
			fileContent = fileContent.replaceAll("</tokenized>", "");
			out.println("<?xml version=\"1.0\" encoding=\""+ encoding+"\"?>");
			out.println("<tokenized>");
			counter_docs = 1;
			out.println("<dataSet name=\"" + name + "\"" +" time=\"" + dataSetTime + "\"" + " language=\"" + language + "\">");
			for (int i = 0; i < hs.getNewsList().size(); i++) {
				out.println("<text tid=\"" + counter_docs + "\">");
				out.println(hs.getNewsList().get(i));
				out.println("</text>");
			}			
			numberOfTexts = hs.getNewsList().size();
		}
		else{
			/* If file is created --> appends */
			int dataSetIndex = fileContent.indexOf("</dataSet>");
			out.println(fileContent.substring(0, dataSetIndex));
			counter_docs++;
			for (int i = 0; i < hs.getNewsList().size(); i++) {		
				out.println("<text tid=\"" + counter_docs + "\">");
				out.println(hs.getNewsList().get(i));
				out.println("</text>");
			}
			numberOfTexts += hs.getNewsList().size();
		}		
		out.println("</dataSet>");
		out.println("</tokenized>");
		out.flush();
		out.close();
		/* Setting isFileMade to true */
		isFileMade = true;
	}
	
	/**
	 * Metoden leser innholdet i fila og lagrer det i en string
	 * @param filename File
	 * @return String
	 * @throws java.io.IOException
	 */
	public String readFile(File filename) throws IOException{
		
		BufferedReader in = new BufferedReader(new FileReader(filename));
		String file = "";
		String temp;
		
		while((temp = in.readLine())!=null){	
			file+=temp;	
			file+="\n";	//Fixes the layout in tokenized.xml!
		}
		return file;
	}
	
	/**
	 * Method that returns number of texts (news) in the collection
	 * @return numberOfTexts
	 */
	public String getNumberOfTexts(){
		return ""+numberOfTexts;
	}
	
	/**
	 * Main method for testing.
	 * @param args
	 */
	public static void main(String[] args){
		System.out.println("Before the second config file...");
		String fileName2 = "config.xml";
		File readFile2 = new File(fileName2);
		XMLParseren xmlpen2 = new XMLParseren(readFile2, true);
		System.out.println("After the second config file...");
	}
	
}