package datamodel;

import java.util.ArrayList;
import java.util.StringTokenizer;

/**
 * Class representing a text. This entity contains the text and the specific 
 * terms in it. 
 * 
 * @author Kjell-Inge Skogstad and Trond ivind Eriksen
 */
public class Text {

    private int tid;
    private ArrayList<Term> terms;
    private String text;
    private int cluster;
    private int noOfTokens;


    /**
     * Constructor creating a new text. The specific terms are not produced and
     * therefore only a empty list of terms are available.
     * 
     * @param tid The timestamp
     * @param text The text
     */
    public Text(int tid, String text) {
        this.tid = tid;
        this.text = text;
        StringTokenizer st = new StringTokenizer(text);
        noOfTokens =st.countTokens();
        terms = new ArrayList<Term>();
    }

    public void changeText(String text){
        this.text=text;
    }

    public void appendText(String text){
        if(this.text.equals("")){
            this.text+=text;
        } else {
            this.text+=" "+text;
        }
    }

    public int getNoOfTokens(){
        return noOfTokens;
    }
    /**
     * Method for adding a term to the text model. If term already exists,
     * the new frequency is added. Weights are set to new weight.
     * 
     * @param term The term
     */
    public void addTerm(String term, int frequency, double weight) {
        int termIndex = findTerm(term);

        /* Testing term index to find if the term is present already */
        if (termIndex == -1) {
            terms.add(new Term(term, frequency, weight));
        } else {
           Term oldTerm = (Term)terms.get(termIndex);
           int oldFrequency = oldTerm.getFrequency();

           oldTerm.setFrequency(oldFrequency + frequency);
           oldTerm.setWeight(weight);
        }
    }

    /**
     * Method for removing a term from the text.
     * 
     * @param term The term to be removed
     * @return The term instance
     */
    public Term removeTerm(Term term) {
        return (Term)terms.remove(findTerm(term.toString()));
    }

    /**
     * Method used to reset the terms. Empties the list of terms.
     */
    public void removeAllTerms() {
        terms = new ArrayList<Term>();
    }
    /**
     * Method for finding the term index by its string representation.
     * 
     * @param term The string representation
     * @return The index
     */
    private int findTerm(String term) {
        int index = -1;

        for (int i = 0; i < terms.size(); i++) {
            if (term.equals(((Term)terms.get(i)).toString())) {
                index = i;
                break;
            }
        }
        return index;
    }

    /**
     * Method returning the timestamp.
     * 
     * @return The timestamp
     */
    public int getTid() {
        return tid;
    }

    /**
     * Method returning all terms in the text.
     * 
     * @return Terms in the text
     */
    public ArrayList getTerms() {
        return terms;
    }

    /**
     * Method for getting wether terms are extracted or not.
     * 
     * @return True if has terms
     */
    public boolean hasTerms() {
        return !terms.isEmpty();
    }

    /**
     * Method returning a printable report of the terms in the text. The report
     * includes information about frequency and weight.
     * 
     * @return The report
     */
    public String getPrintableTerms() {
        String string = "";

        if (terms.size() > 0) {
            string = "(" + ((Term)terms.get(0)).toString() + ", " +
                       ((Term)terms.get(0)).getFrequency() + ", " +
                       ((Term)terms.get(0)).getWeight() + ")";
            for (int i = 1; i < terms.size(); i++) {
                string += " (" + ((Term)terms.get(i)).toString() + ", " +
                            ((Term)terms.get(i)).getFrequency() + ", " +
                            ((Term)terms.get(i)).getWeight() + ")";
            }
        }

        return string;
    }

    /**
     * Method returning a string representation of the text.
     */
    public String toString() {
        return text;
    }

    public int getCluster() {
        return cluster;
    }

    public void setCluster(int cluster) {
        this.cluster = cluster;
    }


}
