package datapreparation.collocation;

import datamodel.DataSet;
import datamodel.Text;
import datamodel.Term;

import java.util.Hashtable;
import java.util.ArrayList;
import java.text.BreakIterator;

/**
 * Class implementing a term collection. A term collection consists of a list
 * of terms with total frequence (sum of frequence in all texts) and the number
 * of text the term is present in. Each term also has a list of which texts it is present in.
 *
 * @author Ole Kristian Fivelstad
 */
public class Collection {

    private Hashtable<String, TermDetails> hashtable;
    private int totalNumberOfTokens;

    /**
     * Constructor creating a new hashtable for storing terms.
     */
    public Collection() {
        hashtable = new Hashtable<String, TermDetails>();
    }

    /**
     * Method for filling the collection. A dataSet is given and parsed so that
     * all individual terms are stored in the collection.
     *
     * @param set The dataset
     */
    public void fillCollection(DataSet set) {
        for (int i = 0; i < set.getNofTexts(); i++) {
            Text text = (Text)set.getTexts().get(i);
            totalNumberOfTokens+=text.getNoOfTokens();
            handleText((Text)set.getTexts().get(i), i);
        }
    }

    /**
     * Method for getting the total number of tokens
     * in the dataset.
     *
     * @return The total number
     */
    public int getTotalNumberOfTokens(){
        return totalNumberOfTokens;
    }

    /**
     * Method for dealing with all terms in a given text. Inserting the terms
     * into the collection.
     *
     * @param text The text
     */
    private void handleText(Text text, int textId) {

        ArrayList terms = text.getTerms();
        Term term;
        TermDetails td;

        for (int i = 0; i < terms.size(); i++) {
            term = (Term)terms.get(i);


            String termString = removeChars(term.toString());
            if (hashtable.containsKey(termString.toLowerCase())) {
                td = hashtable.get(termString.toLowerCase());
                td.addToFreq(term.getFrequency());
                td.addToDocuments(textId);
            } else {
                td = new TermDetails();
                td.addToFreq(term.getFrequency());
                td.addToDocuments(textId);
                hashtable.put(termString.toLowerCase(), td);
            }           
        }
    }

    /**
     * Method for removing specific characters from a word.
     *
     * @param word The word
     * @return The cleaned word
     */
    public String removeChars(String word){
        word=word.replaceAll("\\.","");
        word=word.replaceAll("\\,","");
        word=word.replaceAll("\\'","");
        word=word.replaceAll("\\!","");
        word=word.replaceAll("\\?","");
        word=word.replaceAll("\\(","");
        word=word.replaceAll("\\)","");
        word=word.replaceAll("\\;","");
        word=word.replaceAll("\\:","");
        return word;
    }

    /**
     * Method for getting the total frequency of a given term.
     *
     * @param term The term
     * @return The frequency
     */
    public int getFrequency(String term) {
        return hashtable.get(term.toLowerCase()).freq;
    }

    /**
     * Method for getting the number of texts the given term is present in.
     *
     * @param term The term
     * @return The number of texts
     */
    public int getNofTexts(String term) {
        return hashtable.get(term.toLowerCase()).nofTexts;
    }

    /**
     * Method for getting which documents a term appears in.
     *
     * @param term The term
     * @return The documents it appears in.
     */
    public ArrayList getDocuments(String term) {
        return hashtable.get(term.toLowerCase()).documentIds;
    }

    /**
     * Method for calculating the number of possible collocations in a text
     *
     * @param text The text
     * @return Number of possible collocations
     */
    public int possibleCollocations(String text){
        int possibleCollocations=0;
        ArrayList<String> sentences = new ArrayList<String>();
        BreakIterator bIterator = BreakIterator.getSentenceInstance();
        bIterator.setText(text);
        int start = bIterator.first();
        for (int end = bIterator.next();
             end != BreakIterator.DONE;
             start = end, end = bIterator.next()) {
            int startOfTag = text.indexOf("/", end);
            if(startOfTag!=-1){
                int endOfTag = text.indexOf(" ", startOfTag);
                end=endOfTag;
            }
            if(start>=0&&end>start){
                if(!text.substring(start,end).trim().equals("")){
                    sentences.add(text.substring(start,end).trim());
                }
            }
        }
        for(int i=0; i<sentences.size();i++){
            String sentence = sentences.get(i);
            String[] words = sentence.split(" ");
            int counter=0;
            for(int j=0;j<words.length;j++){
                String tempWord = words[j];
                int tagIndex = tempWord.indexOf("/");
                if(tagIndex!=-1){
                    tempWord=tempWord.substring(0, tagIndex);
                    if(tempWord.endsWith(",") || tempWord.startsWith("(") || tempWord.endsWith(")")){
                        possibleCollocations+=counter;
                        counter=0;
                    } else{
                        counter++;
                    }

                }
            }
            possibleCollocations+=(counter-1);
        }
        return possibleCollocations;
    }


    /**
     * Internal class for representing the details, frequency and number
     * of texts, of a term.
     *
     * @author Kjell-Inge Skogstad and Trond ivind Eriksen
     */
    private class TermDetails {

        private int freq;
        private int nofTexts;
        private ArrayList documentIds;
        /**
         * Constructor creating a new detail container.
         */
        public TermDetails() {
            freq = 0;
            nofTexts = 0;
            documentIds = new ArrayList();
        }

        /**
         * Method for adding to the frequency. Also updates the number of texts
         * the term is present in.
         *
         * @param add Frequency to add
         */
        public void addToFreq(int add) {
            freq += add;
            nofTexts++;
        }

        /**
         * Method for adding to the list of documents it appears in.
         *
         * @param textId Id of the text the term appears in.
         */
        public void addToDocuments(int textId){
            documentIds.add(textId);
        }
    }
}
