package datapreparation.collocation;

import datapreparation.TextOperation;
import datapreparation.stopword.EnglishStopwords;
import datamodel.DataSet;
import datamodel.Text;

import java.util.ArrayList;
import java.util.StringTokenizer;
import java.text.BreakIterator;

import gui.MainFrame;
import net.didion.jwnl.JWNLException;
import net.didion.jwnl.dictionary.Dictionary;
import net.didion.jwnl.data.IndexWord;
import net.didion.jwnl.data.POS;
import control.OperationProperty;
import control.Globals;

/**
 * Class representing the collocation extraction operation. This operation tries to find
 * collocations of two nouns. In addition, proper nouns and proper noun groups are extracted.
 * The user may also specify the extraction of verbs, adjectives, adverbs and numbers.
 *
 * @author Ole Kristian Fivelstad
 *
 */
public class Collocation extends TextOperation {

    private Collection collection;
    private int totalNoOfTokens;
    private double significanceValue=3.841;
    private ArrayList<String> dividers;
    private EnglishStopwords stopwords;


    //variables indicating which word classes to extract
    private boolean adjectives=true;    
    private boolean verbs=true;
    private boolean adverbs=false;
    private boolean numbers=false;

    private int collocations=0;

    private ArrayList originalTexts;

    /**
     * Constructor for the collocation operation. Initializes a vector of possible
     * sentence dividers. For example,
     * , ) ( and so on.
     * An english stopwords remover is also initialized.
     */
    public Collocation(){
        dividers = new ArrayList<String>();
        dividers.add("'");
        dividers.add(",");
        dividers.add("");
        dividers.add("(");
        dividers.add(")");
        dividers.add("");
        dividers.add(":");
        dividers.add(";");
        dividers.add("\"");
        stopwords=new EnglishStopwords();
    }

    /**
     * Method for performing the operation
     *
     * @param dataSet  The dataset being used
     */
    public void performOperation(DataSet dataSet) {
        if(dataSet.isPartOfSpeechPerformed()){
            collection = new Collection();
            int numberOfTexts = dataSet.getNofTexts();
            MainFrame mainFrame = MainFrame.getInstance();
            collection.fillCollection(dataSet);
            totalNoOfTokens=collection.getTotalNumberOfTokens();
            ArrayList texts = dataSet.getTexts();
            originalTexts = (ArrayList)dataSet.getTexts().clone();
            for (int i = 0; i < numberOfTexts; i++) {
                texts.add(i, extractTermsFromText((Text)texts.remove(i)));
                int progress = (int)((double)(i + 1)/(double)numberOfTexts * 100);
                if(Globals.showGui){
                    mainFrame.updateOperationProgress(progress);
                }
            }
            System.out.println("Total number of collocations: "+collocations);
            dataSet.setCollocationPerformed(true);
        } else{
            System.out.println("Part of speech must be performed before collocation...");
        }
    }

    /**
     * Method for extracting terms from a specific text
     *
     * @param text The text being processed
     * @return The new text
     */
    public Text extractTermsFromText(Text text){
        String stringText = text.toString();
        Text newText = new Text(text.getTid(), stringText);
        newText.changeText("");
        ArrayList sentences = findSentences(stringText);
        for(int i=0;i<sentences.size();i++){
            String sentence = sentences.get(i).toString();
            extractTermsFromSentence(sentence, newText);
        }
        return newText;

    }

    /**
     * Method for looking up the lemma of a noun in WordNet.
     *
     * @param word The word whose lemma is wanted
     * @return The lemma of the word
     */
    public String findNounLemma(String word){
        String lemma="";
        String temp = word.substring(0,word.indexOf("/NN"));
        temp = removeChars(temp);
        try {
            IndexWord indexWord = Dictionary.getInstance().getMorphologicalProcessor().
                    lookupBaseForm(POS.NOUN, temp.toLowerCase());
            if(indexWord!=null){
                lemma = indexWord.getLemma()+"/NN";
            }
        } catch (JWNLException e) {
            e.printStackTrace();
        }
        return lemma;
    }

    /**
     * Method for extracting terms from a sentence.
     *
     * @param sentence The sentence being processed
     * @param newText The text the sentence belongs to
     *
     */
    public void extractTermsFromSentence(String sentence, Text newText){

        String newStringText="";

        String words[] = sentence.split(" ");
        String word="";
        String word2="";
        String word3="";
        String word4="";

        for(int i=0; i<words.length;i++){
            word = words[i];
            word2="";
            word3="";
            word4="";


            //Looking for various patterns and word-classes in the tagged text.

            //adverbs
            if(word.indexOf("/RB")!=-1&&adverbs){
                String adverb=word;
                newText.addTerm(removeChars(adverb),1,0);
                newStringText+=" "+removeChars(adverb);
            }
            //adjectives
            else if(word.indexOf("/JJ")!=-1&&adjectives){
                String adjective="";
                try {
                    String adj = word.substring(0, word.indexOf("/JJ"));
                    adj=removeChars(adj);
                    if(!adj.contains("-")){
                        IndexWord indexWord = Dictionary.getInstance().getMorphologicalProcessor().
                                lookupBaseForm(POS.ADJECTIVE, adj.toLowerCase());
                        if(indexWord!=null){
                            adjective=stopwords.removeStopwords(adjective);
                            if(!adjective.equals("")){
                                adjective = indexWord.getLemma()+"/JJ";
                            }
                        }
                    } else {
                        adjective = adj+"/JJ";
                    }
                } catch (JWNLException e) {
                    e.printStackTrace();
                }
                if(!adjective.equals("")){
                    newText.addTerm(adjective,1,0);
                    newStringText+=" "+adjective;
                }
            }
            //verbs
            else if(word.indexOf("/VB")!=-1&&verbs){
                String temp = word.substring(0,word.indexOf("/VB"));
                temp = removeChars(temp);
                String verb = "";
                try {
                    IndexWord indexWord = Dictionary.getInstance().getMorphologicalProcessor().
                            lookupBaseForm(POS.VERB, temp.toLowerCase());
                    if(indexWord!=null){
                        verb=stopwords.removeStopwords(indexWord.getLemma().trim());
                        if(!verb.trim().equals("")){
                            verb = indexWord.getLemma()+"/VB";
                            newText.addTerm(verb,1,0);
                            newStringText+=" "+verb;
                        }
                    }
                } catch (JWNLException e) {
                    e.printStackTrace();
                }
            }
            //numbers
            else if(word.indexOf("/CD")!=-1&&numbers){
                String numbers = word;
                newStringText+=" "+numbers;
                newText.addTerm(numbers,1,0);
            }
            //nouns
            else if(word.indexOf("/NN")!=-1){
                //proper nouns
                if(word.indexOf("/NNP")!=-1){
                    word=removeChars(word);
                    if(i+1<words.length){
                        word2=words[i+1];
                    }
                    if(i+2<words.length){
                        word3=words[i+2];
                    }
                    if(i+3<words.length){
                        word4=words[i+3];
                    }
                    word2=removeChars(word2);
                    word3=removeChars(word3);
                    word4=removeChars(word4);

                    //Trying to find various patterns of proper nouns

                    if(word2.indexOf("/IN")!=-1&&!word2.contains("with")&&word3.indexOf("/NNP")!=-1){
                        String properNounOne = word.substring(0,word.indexOf("/NNP"));
                        String inflection = word2.substring(0, word2.indexOf("/IN"));
                        String properNounTwo = word3.substring(0, word3.indexOf("/NNP"));
                        String compoundNoun = properNounOne+"_"+inflection+"_"+properNounTwo;
                        // NNP IN NNP, e.g. Ministry of Defence
                        if(word4.indexOf("/NNP")==-1){
                            newText.addTerm(compoundNoun+"/NNP",1,0);
                            newStringText+=" "+compoundNoun+"/NNP";
                            i+=3;
                        }
                        // NNP IN NNP NNP , e.g. Council on Foreign Relations
                        else{
                            String properNounThree = word4.substring(0,word4.indexOf("/NNP"));
                            compoundNoun=compoundNoun+"_"+properNounThree+"/NNP";
                            newText.addTerm(compoundNoun,1,0);
                            newStringText+=" "+compoundNoun;
                            i+=4;
                        }
                    }
                    //finding groups(up to 3) of proper nouns, e.g. George Bush, President Vladimir Putin
                    else if(word2.indexOf("/NNP")!=-1){

                        if(word3.indexOf("/NNP")!=-1){
                            //Group of 3 proper nouns
                            String wordOne = word.substring(0,word.indexOf("/NNP"));
                            String wordTwo = word2.substring(0,word2.indexOf("/NNP"));
                            String wordThree = word3.substring(0,word3.indexOf("/NNP"));
                            String compound = wordOne+"_"+wordTwo+"_"+wordThree+"/NNP";
                            newText.addTerm(compound,1,0);
                            newStringText+=" "+compound;
                            i+=2;
                            compound="";
                        }
                        else{
                            //Group of 2 proper nouns
                            String wordOne = word.substring(0,word.indexOf("/NNP"));
                            String wordTwo = word2.substring(0,word2.indexOf("/NNP"));
                            String compound = wordOne+"_"+wordTwo+"/NNP";
                            newText.addTerm(compound,1,0);
                            newStringText+=" "+compound;
                            i+=1;
                            compound="";
                        }
                    }
                    else{
                        //Only one proper noun
                        String properNoun = removeChars(word);//word.substring(0,word.indexOf("/NNP"));
                        newStringText+=" "+properNoun;
                        newText.addTerm(properNoun,1,0);
                    }
                }
                else{
                    //common nouns
                    boolean firstWordIsValid=wordIsValid(word);
                    boolean secondWordIsValid=false;
                    //boolean thirdWordIsValid=false;
                    if(i+1<words.length){
                        word2=words[i+1];
                        if(word2.indexOf("/NN")!=-1&&word2.indexOf("/NNP")==-1){
                            secondWordIsValid=wordIsValid(word2);
                        }
                    }

                    if(i+2<words.length){
                        word3=words[i+2];
                        //if(word3.indexOf("/NN")!=-1&&word3.indexOf("/NNP")==-1){
                        //    thirdWordIsValid=wordIsValid(word3);
                        //}
                    }

                    //checking to see if two common nouns appear together in the text
                    //without any dividers between them
                    if(firstWordIsValid && word.indexOf("/NNS")==-1){
                        if(secondWordIsValid&&!firstCharIsDivider(word2)&&!wordEndsWithComma(word)
                                &&!word.equals(word2)){

                            //checking if it is likely to be a collocation
                            boolean collocation = calculateCollocation(word, word2);
                            if(collocation){

                                /*
                                if(thirdWordIsValid&&!firstCharIsDivider(word3)&&!wordEndsWithComma(word2)){
                                    boolean collocat = calculateCollocation(word2, word3);
                                    if(collocat){
                                        System.out.println(word+" "+word2+" "+word3);
                                    }
                                }*/

                                String tempWord = removeChars(word.substring(0, word.indexOf("/")));
                                String tempWord2 = removeChars(word2.substring(0, word2.indexOf("/")));
                                String coll = generateCollocation(tempWord, tempWord2);                                        
                                newText.addTerm(coll, 1, 0);
                                newStringText+=" "+coll;
                                i+=1;
                                coll="";
                                collocations++;
                            } else {
                                //Not a collocation, adding the first noun
                                String noun = findNounLemma(word);
                                if(!noun.equals("")){
                                    newText.addTerm(noun,1,0);
                                    newStringText+=" "+noun;
                                }
                            }
                        } else{
                            //A single noun
                            String noun = findNounLemma(word);
                            if(!noun.equals("")){
                                newText.addTerm(noun,1,0);
                                newStringText+=" "+noun;
                            }
                        }

                    }
                    //plural common nouns
                    else if(firstWordIsValid && word.indexOf("/NNS")!=-1){
                        String noun = findNounLemma(word);
                        if(!noun.equals("")){
                            newText.addTerm(noun,1,0);
                            newStringText+=" "+noun;
                        }
                    } else{
                        //Do nothing...
                        //Term has been incorrectly tagged by the tagger.
                    }
                }
            }

        }
        newText.appendText(newStringText.trim());
    }

    /**
     * Method for generating a collocation of two words.
     * This is done by putting _ between the two words.
     * E.g. interest_rate
     *
     * @param firstWord The first word in the collocation
     * @param secondWord The second word in the collocation
     * @return The collocation string
     */
    public String generateCollocation(String firstWord, String secondWord){
        String collocation="";
        String partOne = "";
        String partTwo = "";
        try{
            if(firstWord.contains("-")){
                partOne = firstWord.replaceAll("-","_");                
                collocation+=partOne;
            } else {
            IndexWord firstLemma = Dictionary.getInstance().getMorphologicalProcessor().
                    lookupBaseForm(POS.NOUN, firstWord.toLowerCase());
                if(firstLemma!=null){
                    partOne = firstLemma.getLemma();
                    collocation+=partOne;
                }
            }
            if(secondWord.contains("-")){
                partTwo = secondWord.replaceAll("-","_");
                collocation+="_"+partTwo+"/COLL";
            } else {
            IndexWord secondLemma = Dictionary.getInstance().getMorphologicalProcessor().
                    lookupBaseForm(POS.NOUN, secondWord.toLowerCase());
                if(secondLemma!=null){
                    partTwo = secondLemma.getLemma();
                    collocation+="_"+partTwo+"/COLL";
                }
            }
        } catch (Exception e){
            e.printStackTrace();
        }

        if(collocation.equals("")){
            collocation = firstWord+"_"+secondWord+"/COLL";
        }
        return collocation;
    }

    /**
     * Method for checking whether a word is valid.
     * This is done by both looking it up in WordNet, and checking
     * if it is a stop word.
     *
     * @param word The word being checked
     * @return The result
     */
    public boolean wordIsValid(String word){
       word = word.substring(0, word.indexOf("/"));
       word = removeChars(word);
       word = word.toLowerCase();
        try {
            IndexWord temp = Dictionary.getInstance().getMorphologicalProcessor()
                    .lookupBaseForm(POS.NOUN, word);
            if(temp==null){
                return false;
            } else{
                if(wordIsStopword(temp.getLemma()) || temp.getLemma().length()<2){
                    return false;
                } else {
                    return true;
                }
            }
        } catch (JWNLException e) {
            System.out.println("Problem accessing WordNet data...");
            return false;
        }
    }

    /**
     * Method for checking whether a word is a stopword.
     *
     * @param word The word being checked.
     * @return The result
     */
    public boolean wordIsStopword(String word){
        word = stopwords.removeStopwords(word);
        if((word.trim()).equals("")){
            return true;
        } else{
            return false;
        }
    }

    /**
     * Method for performing the actual calculation too see
     * if two words form a collocation.
     *
     * @param firstWord The first word
     * @param secondWord The second word
     * @return The result
     */
    public boolean calculateCollocation(String firstWord, String secondWord){
        boolean result = false;

        firstWord = removeChars(firstWord);
        secondWord = removeChars(secondWord);

        //Getting the frequencies of the words in the collection
        int w1 = collection.getFrequency(firstWord);
        int w2 = collection.getFrequency(secondWord);
        int w1w2 = 0;
        int neither=0;

        //Finding how frequent they appear together
        if(w1==1&&w2==1){
            w1w2=1;
        } else{
            ArrayList documentsOne = collection.getDocuments(firstWord);
            ArrayList documentsTwo = collection.getDocuments(secondWord);
            ArrayList overlap = findOverlap(documentsOne, documentsTwo);
            for(int i=0;i<overlap.size();i++){
                int textIndex = Integer.parseInt(overlap.get(i).toString());
                Text text = (Text)originalTexts.get(textIndex);
                String textString = text.toString();
                ArrayList terms = new ArrayList();
                StringTokenizer st = new StringTokenizer(textString);
                while(st.hasMoreTokens()){
                    String temp = st.nextToken();
                    if(temp.indexOf("/CD")==-1){
                        temp=removeChars(temp);
                    }
                    terms.add(temp);
                }
                for(int j=0;j<terms.size();j++){
                    if(terms.get(j).toString().equals(firstWord)){
                        if(j<terms.size()-1){
                            if(terms.get(j+1).toString().equals(secondWord)){
                                w1w2++;
                                j++;
                            }
                        }
                    }
                }
            }
        }

        //Calculating the chi-square value according to the forumula presented in the report
        w1=w1-w1w2;
        w2=w2-w1w2;
        neither=totalNoOfTokens-w1-w2-w1w2;
        double numerator = (w1w2*neither)-(w1*w2);
        numerator=(numerator*numerator);

        //The calculation is split up, since the numerator can become very high
        //and cause integer overflow.
        numerator = numerator / (w1w2+w2);
        numerator = numerator / (w1w2+w1);
        numerator = numerator / (w2+neither);
        numerator = numerator / (w1+neither);

        double chiSquare = numerator;

        chiSquare = chiSquare*totalNoOfTokens;

        //Checking to see if the result is higher than the significance value.
        if(chiSquare>=significanceValue){
            result=true;
        }
        return result;
    }

    /**
     * Method for finding which documents the two terms appear together in
     * @param one The documents of word one
     * @param two The documents of word two
     * @return ArrayList containing the overlap
     */
    public ArrayList findOverlap(ArrayList one, ArrayList two){
        ArrayList overlap = new ArrayList();
        for(int i = 0; i<one.size(); i++){
            if(two.contains(one.get(i))){
                overlap.add(one.get(i));
            }
        }
        return overlap;
    }

    /**
     * Method for finding the sentences in a text.
     * @param text The text
     * @return ArrayList containing the sentences
     */
    public ArrayList findSentences(String text){
        ArrayList<String> sentences = new ArrayList<String>();
        BreakIterator bIterator = BreakIterator.getSentenceInstance();
        bIterator.setText(text);
        int start = bIterator.first();
        for (int end = bIterator.next();
             end != BreakIterator.DONE;
             start = end, end = bIterator.next()) {
            int startOfTag = text.indexOf("/", end);
            if(startOfTag!=-1){
                int endOfTag = text.indexOf(" ", startOfTag);
                end=endOfTag;
            }
            if(start>=0&&end>start){
                sentences.add(text.substring(start,end).trim());
            }
        }
        return sentences;
    }

    /**
     * Method for checking if a word contains on of a list of dividers
     *
     * @param word The word being checked
     * @return The result
     */
    public boolean wordContainsChar(String word){
        for(int i=0; i<dividers.size();i++){
            if(word.contains(dividers.get(i).toString())){
                return true;
            }
        }
        return false;
    }

    /**
     * Method for checking to see if the first character of a word
     * is a divider. E.g. ( or '
     * @param word The word being checked
     * @return The result
     */
    public boolean firstCharIsDivider(String word){
        String substring = word.substring(0,1);
        if(dividers.contains(substring)){
            //System.out.println("First char is divider: "+word);
            return true;
        }

        return false;
    }

    /**
     * Method for checking if the last character in a word is a comma.
     *
     * @param word The word being checked
     * @return The result
     */
    public boolean wordEndsWithComma(String word){
        int index = word.indexOf("/");
        index-=1;
        String substring = word.substring(index, index+1);
        if(substring.equals(",")){
            //System.out.println("Word ends with comma (oppramsing) : "+word);
            return true;
        }
        return false;
    }

    /**
     * Method for removing specific characters from a word.
     *
     * @param word The word
     * @return The new word
     */
    public String removeChars(String word){
        word=word.replaceAll("\\.","");
        word=word.replaceAll("\\,","");
        word=word.replaceAll("\\'","");
        word=word.replaceAll("\\`","");
        word=word.replaceAll("\\!","");
        word=word.replaceAll("\\?","");
        word=word.replaceAll("\\(","");
        word=word.replaceAll("\\)","");
        word=word.replaceAll("\\;","");
        word=word.replaceAll("\\:","");
        return word;
    }
    /**
     * Method for getting the properties
     *
     * @return properties
     */

    public ArrayList getProperties() {
        ArrayList properties = new ArrayList();

        properties.add(new OperationProperty("Keep verbs",verbs+""));
        properties.add(new OperationProperty("Keep adverbs",adverbs+""));
        properties.add(new OperationProperty("Keep adjectives",adjectives+""));
        properties.add(new OperationProperty("Keep numbers",numbers+""));

        return properties;
    }

    /**
     * Method for setting the properties
     *
     * @param properties
     */
    public void setProperties(ArrayList properties) {
        OperationProperty p;

        for (int i = 0; i < properties.size(); i++) {
            p = (OperationProperty)properties.get(i);
            try {
                if (p.getName().equals("Keep verbs")) {
                    verbs = Boolean.parseBoolean(p.getValue());
                } else if (p.getName().equals("Keep adverbs")) {
                    adverbs = Boolean.parseBoolean(p.getValue());
                } else if (p.getName().equals("Keep adjectives")) {
                    adjectives = Boolean.parseBoolean(p.getValue());
                } else if (p.getName().equals("Keep numbers")) {
                    numbers = Boolean.parseBoolean(p.getValue());
                }
            } catch (Exception e) {
                System.out.println("Error parsing property " + p.getName());
            }
        }
    }
}
