package datapreparation.concept;

import datapreparation.TextOperation;
import datamodel.DataSet;
import datamodel.Text;
import datamodel.Term;

import java.util.ArrayList;

import net.didion.jwnl.JWNLException;
import net.didion.jwnl.data.*;
import net.didion.jwnl.data.list.PointerTargetNodeList;
import net.didion.jwnl.data.list.PointerTargetNode;
import net.didion.jwnl.data.relationship.RelationshipFinder;
import net.didion.jwnl.data.relationship.RelationshipList;
import net.didion.jwnl.data.relationship.AsymmetricRelationship;
import net.didion.jwnl.dictionary.Dictionary;
import control.OperationProperty;
import control.Globals;
import gui.MainFrame;

/**
 * Class implementing the concept extraction operation. The operation finds by concepts by
 * both finding common parents between words, and by checking the category-relation of the words.
 *
 * @author Ole Kristian
 *
 */
public class Concept extends TextOperation {

    private ArrayList candidates;
    private int numberOfTerms;
    private int conceptsFound=0;

    //Some parameters for the operation
    private boolean addParentNode=false;
    private int maximumDistanceInHierarchy = 5;
    private boolean keepOriginalTerms = false;


    /**
     * Method for performing the operation.
     *
     * @param dataSet The dataset being used
     */
    public void performOperation(DataSet dataSet) {
        if(dataSet.isWsdPerformed()){
            ArrayList<Text> texts = dataSet.getTexts();
            int numberOfTexts = dataSet.getNofTexts();

            for (int i = 0; i < numberOfTexts; i++) {
                candidates=new ArrayList();
                texts.add(i, extractConcepts((Text)texts.remove(i)));
                int progress = (int)((double)(i + 1)/(double)numberOfTexts * 100);
                if(Globals.showGui){
                    MainFrame.getInstance().updateOperationProgress(progress);
                }
                candidates=null;
            }
            System.out.println("Antall konsepter funnet: "+conceptsFound);
        } else{
            System.out.println("WSD must be performed before concept extraction");
        }
    }

    /**
     * Method for finding the terms which have been disambiguated.
     * These are candidates for concept extraction.
     *
     * @param terms The terms in the text.
     */
    public void initializeCandidates(ArrayList terms){
        for(int i=0; i<numberOfTerms;i++){
            Term term = ((Term)terms.get(i));
            String termString = term.toString();
            if(termString.contains("#")){
                candidates.add(termString);
            }
        }
    }

    /**
     * Method for extracting concepts from a text.
     * @param text The text
     * @return The new text
     */
    public Text extractConcepts(Text text){
        String stringText = text.toString();
        Text newText = new Text(text.getTid(), stringText);
        ArrayList terms = text.getTerms();
        numberOfTerms=terms.size();
        initializeCandidates(terms);
        ArrayList concepts=new ArrayList();
        for(int i=0; i<terms.size(); i++){
            Term term = ((Term)terms.get(i));
            String termString = term.toString();
            //All terms containing # have been disambiguated, and are candidates for concept
            //extraction.
            if(termString.contains("#")){
                //find concepts for the term
                concepts=findConcepts(termString);

                if(concepts.size()==0){
                    //Term did not have any concepts
                    Synset synset = findSynset(termString);
                    if(synset==null){
                        newText.addTerm(termString, term.getFrequency(), term.getWeight());
                    } else {
                        String word = synset.getWord(0).getLemma();
                        int sense = findSense(synset.getOffset(), word);
                        if(synset.getWordsSize()>1){
                            word=word+"/"+synset.getWord(1).getLemma();
                        }
                        newText.addTerm(word+"#"+sense, term.getFrequency(), term.getWeight());
                    }
                    //adding parent node of term if this is wanted by the user
                    if(addParentNode){
                        ArrayList parents = findParent(termString);
                        for(int j=0; j<parents.size();j++){
                            newText.addTerm(parents.get(j).toString(), 1, 0);
                        }
                    }
                } else{
                    //concepts are found, these are added to the new text.
                    for(int j=0; j<concepts.size(); j++){
                        conceptsFound++;
                        newText.addTerm((String)concepts.get(j), 1, 0);
                    }
                    //if the original terms should be kept,
                    // they are added to the new text as well
                    if(keepOriginalTerms){
                        Synset synset = findSynset(termString);
                        if(synset==null){
                            newText.addTerm(termString, term.getFrequency(), term.getWeight());
                        } else {
                            String word = synset.getWord(0).getLemma();
                            int sense = findSense(synset.getOffset(), word);
                            if(synset.getWordsSize()>1){
                                word=word+"/"+synset.getWord(1).getLemma();
                            }
                            newText.addTerm(word+"#"+sense, term.getFrequency(), term.getWeight());
                        }
                    }
                }
            //Collocations and proper nouns are added to the text.
            //Along with any other word class previously extracted
            } else if(termString.contains("/COLL")) {
                newText.addTerm(termString, term.getFrequency(), term.getWeight());
            } else if(termString.contains("/NNP")) {
                newText.addTerm(termString, term.getFrequency(), term.getWeight());
            } else {
                newText.addTerm(termString, term.getFrequency(), term.getWeight());
            }
        }
        return newText;
    }

    /**
     * Method for getting the synset of a term.
     *
     * @param term The term
     * @return The synset
     */
    public Synset findSynset(String term){
        String temp = term.substring(0, term.indexOf("#"));
        try {
            IndexWord candidate = Dictionary.getInstance().getIndexWord(POS.NOUN, temp.toLowerCase());
            int sense = Integer.
                    parseInt(term.substring(term.indexOf("#")+1, term.length()));
            Synset synset = candidate.getSense(sense);
            return synset;
        } catch (Exception e){

        }
        return null;
    }

    /**
     * Method for finding the parent node of a word in WordNet
     * @param termString The word
     * @return List containing all parent nodes.
     */
    public ArrayList findParent(String termString){
        ArrayList parent=new ArrayList();
        int sense = Integer.parseInt(termString.substring(termString.indexOf("#")+1, termString.length()));
        String lemma = termString.substring(0, termString.indexOf("#"));
        try {
            IndexWord word = Dictionary.getInstance().getIndexWord(POS.NOUN, lemma);
            Synset synset = word.getSense(sense);
            Pointer[] parents = synset.getPointers(PointerType.HYPERNYM);
            for(int i=0; i<parents.length; i++){
                Pointer pointer = parents[i];
                Synset parentSynset = pointer.getTargetSynset();
                int parentSense = findSense(parentSynset.getOffset(), parentSynset.getWord(0).getLemma());
                String parentWord = parentSynset.getWord(0).getLemma();
                if(parentSynset.getWordsSize()>1){
                    parentWord=parentWord+"/"+parentSynset.getWord(1).getLemma();
                }
                parent.add(parentWord+"#"+parentSense);
            }

        } catch (JWNLException e) {
            e.printStackTrace();
        }

        return parent;
    }

    /**
     * Method for finding the common parent between two synsets.
     *
     * @param first The first synset
     * @param second The second synset
     * @return The common parent
     */
    public String findCommonParent(Synset first, Synset second){
        String commonParent="";
        try{
            RelationshipFinder rf = RelationshipFinder.getInstance();
            RelationshipList rl = rf.findRelationships(first, second, PointerType.HYPERNYM);
            if(rl.size()>0){
                AsymmetricRelationship relationship = (AsymmetricRelationship)rl.getShallowest();
                if (relationship!=null){
                    if(relationship.getSize()<=maximumDistanceInHierarchy){
                        int parentIndex = relationship.getCommonParentIndex();
                        PointerTargetNodeList ptnl = relationship.getNodeList();
                        PointerTargetNode ptn = (PointerTargetNode)ptnl.get(parentIndex);
                        Synset parent = ptn.getSynset();
                        long offset = parent.getOffset();
                        String conc = parent.getWord(0).getLemma();
                        int sense = findSense(offset, conc);
                        if(parent.getWordsSize()>2){
                            conc=conc+"/"+parent.getWord(1).getLemma();
                        }
                        commonParent=conc+"#"+sense+"";
                    }
                }
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
        return commonParent;
    }

    /**
     * Method for finding the category of a synset.
     *
     * @param candidate The candidate synset
     * @return List containing the categories of the synset
     */
   public ArrayList<String> findCategory(Synset candidate){
        ArrayList<String> categories = new ArrayList<String>();

        Pointer pointer[] = candidate.getPointers(PointerType.CATEGORY);

        if(pointer.length==0){
            return null;
        }

        for(int i=0; i<pointer.length; i++){
            String category="";
            Pointer pt = pointer[i];
            try {
                Synset targetSynset = pt.getTargetSynset();
                long offset = targetSynset.getOffset();
                String cat = targetSynset.getWord(0).getLemma();
                int sense = findSense(offset, cat);                
                if(targetSynset.getWordsSize()>1){
                    cat=cat+"/"+targetSynset.getWord(1).getLemma();
                }
                category=cat+"#"+sense+"";
                categories.add(category);
            } catch (JWNLException e) {
                e.printStackTrace();
            }
        }
        return categories;
    }

    /**
     * Method for finding the sense of a word.
     * @param offset The offset of the synset the word belongs to in WordNet
     * @param word The word
     * @return The sense
     */
    private int findSense(long offset, String word){
        int sense = -1;
        try {
            IndexWord next = Dictionary.getInstance().getIndexWord(POS.NOUN, word.toLowerCase());
            if(next.getSenseCount()==1) return 1;

            for(int i=1; i<=next.getSenseCount();i++){
                Synset temp = next.getSense(i);
                if(temp.getOffset()==offset){
                    return i;
                }
            }
        } catch (JWNLException e) {
            e.printStackTrace();
        } catch (NullPointerException npe){
            npe.printStackTrace();
        }
        return sense;
    }

    /**
     * Method for finding concepts.
     * This method uses the methods findCategory and findCommonParent.
     * @param candidate The candidate word
     * @return List of concepts
     */
    public ArrayList findConcepts(String candidate){
        candidates.remove(candidate);
        ArrayList concepts = new ArrayList();
        int senseIndex = candidate.indexOf("#")+1;
        int sense = Integer.parseInt(candidate.substring(senseIndex, candidate.length()));
        String word = candidate.substring(0, senseIndex-1);
        try {
            IndexWord candidateWord = Dictionary.getInstance().getIndexWord(POS.NOUN, word.toLowerCase());
            if(candidate==null) return null;
            Synset candidateSynset = candidateWord.getSense(sense);

            ArrayList<String> categories = findCategory(candidateSynset);
            if(categories!=null){
                for(int j=0; j<categories.size(); j++){
                    concepts.add(categories.get(j));                   
                }
            }

            for(int k=0; k<candidates.size(); k++){
                String temp2 = candidates.get(k).toString();
                int senseIndex2 = temp2.indexOf("#")+1;
                int sense2 = Integer.parseInt(temp2.substring(senseIndex2, temp2.length()));
                String word2 = temp2.substring(0, senseIndex2-1);
                IndexWord neighbor = Dictionary.getInstance().getIndexWord(POS.NOUN, word2.toLowerCase());
                if(neighbor==null) break;
                Synset neighborSynset = neighbor.getSense(sense2);
                String concept = findCommonParent(candidateSynset, neighborSynset);
                if(!concept.equals("")){
                    concepts.add(concept);
                }
            }

        } catch(Exception e) {
            e.printStackTrace();
        }

        return concepts;
    }

    /**
     * Method for getting the properties.
     */
    public ArrayList getProperties() {
        ArrayList properties = new ArrayList();

        properties.add(new OperationProperty("Maximum distance in hierarchy",maximumDistanceInHierarchy+""));
        properties.add(new OperationProperty("Keep original terms",keepOriginalTerms+""));
        properties.add(new OperationProperty("Add parent node",addParentNode+""));

        return properties;
    }

    /**
     * Method for setting the properties.
     */
    public void setProperties(ArrayList properties) {
        OperationProperty p;

        for (int i = 0; i < properties.size(); i++) {
            p = (OperationProperty)properties.get(i);
            try {
                if (p.getName().equals("Maximum distance in hierarchy")) {
                    maximumDistanceInHierarchy = Integer.parseInt(p.getValue());
                } else if (p.getName().equals("Keep original terms")) {
                    keepOriginalTerms = Boolean.parseBoolean(p.getValue());
                } else if (p.getName().equals("Add parent node")) {
                    addParentNode = Boolean.parseBoolean(p.getValue());
                }
            } catch (Exception e) {
                System.out.println("Error parsing property " + p.getName());
            }
        }
    }
}
