package datapreparation.filter;

import gui.MainFrame;

import java.util.ArrayList;

import control.OperationProperty;

import datamodel.DataSet;
import datamodel.Term;
import datamodel.Text;
import datapreparation.TextOperation;
import external.QuickSort;

/**
 * Class representing a filtering operation. This operation allows for the 
 * user to filter the extracted terms by weight or by order, specifying a maximum
 * number of terms to keep, and wether to keep the terms read and more and terms 
 * with digits. 
 * 
 * @author Kjell-Inge Skogstad and Trond ivind Eriksen
 */
public class Filter extends TextOperation {

    private int max = 5;
    private boolean byWeight = true;
    private boolean digits = true;
    private boolean read = false;
    private boolean more = false;
    
    /**
     * Method for performing the operation.
     */
    public void performOperation(DataSet dataSet) {
        ArrayList texts = dataSet.getTexts();
        
        if (!dataSet.termExtracted()) {
          //  MainFrame.getInstance().printMessage("Terms are not extracted. " +
            //        "Filter operation aborted");
            return;
        }
        
        for (int i = 0; i < texts.size(); i++) {
            texts.add(i, extractWordsInText((Text)texts.remove(i)));
        }
    }

    /**
     * Method for filtering the terms in a specific text. 
     * 
     * @param text The text
     * @return The filtered text
     */
    private Text extractWordsInText(Text text) {
        ArrayList terms = text.getTerms();
        Text newText = new Text(text.getTid(), text.toString());
        Term term;
        int internalMax = 0;
        
        // Test if legal max
        if (max < 0 || max > terms.size()) {
            internalMax = terms.size();
        } else {
            internalMax = max;
        }
        
        // Test if filtering by weight or by order
        if (byWeight) {
            QuickSort.sort(terms);
            
            //Keeps the max highest weighted terms
            for (int i = 0; i < internalMax; i++) {
                term = (Term)terms.get(i);
                if (keepTerm(term.toString())) {
                    newText.addTerm(term.toString(), term.getFrequency(), term.getWeight());
                }
            }
        } else {
            //Keeps the max first terms
            for (int i = 0; i < internalMax; i++) {
                term = (Term)terms.get(i);
                if (keepTerm(term.toString())) {
                    newText.addTerm(term.toString(), term.getFrequency(), term.getWeight());
                }
            }
        }
        return newText;
    }

    /**
     * Method for testing wether or not the term should be kept according to 
     * specified parameters.
     * 
     * @param term The term
     * @return True if term should be kept
     */
    private boolean keepTerm(String term) {
        if (!term.trim().equals("") && !digits) {
            term = removeDigits(term); 
        }
        
        if (!term.trim().equals("") && !read) {
            term = removeRead(term); 
        }
        
        if (!term.trim().equals("") && !more) {
            term = removeMore(term); 
        }
        
        if (!term.trim().equals("")) {
            return true;
        }
        return false;
    }
    
    /**
     * Methof for removing terms which contain digits.
     * 
     * @param term The term
     * @return The new term (could be empty)
     */
    private String removeDigits(String term) {
        for (int i = 0; i < term.length(); i++) {
            if (Character.isDigit(term.charAt(i))) {
                return "";
            }
        }
        return term;
    }
    
    /**
     * Method for removing terms equal to 'read'.
     * 
     * @param term The term
     * @return The new term (could be empty)
     */
    private String removeRead(String term) {
        if (term.toLowerCase().trim().equals("read")) {
            return "";
        }
        return term;
    }
    
    /**
     * Method for removing terms equal to 'more'.
     * 
     * @param term The term
     * @return The new term (could be empty)
     */
    private String removeMore(String term) {
        if (term.toLowerCase().trim().equals("more")) {
            return "";
        }
        return term;
    }
    
    /**
     * Method returning the properties of this operation.
     */
    public ArrayList getProperties() {
        ArrayList properties = new ArrayList();
        
        properties.add(new OperationProperty("Maximum number of terms", max + ""));
        properties.add(new OperationProperty("Get by weight", 
                new Boolean(byWeight).toString()));
        properties.add(new OperationProperty("Keep words with digits", 
                new Boolean(digits).toString()));
        properties.add(new OperationProperty("Keep the word 'more'", 
                new Boolean(more).toString()));
        properties.add(new OperationProperty("Keep the word 'read'", 
                new Boolean(read).toString()));
        
        return properties;
    }

    /**
     * Method for setting the properties of this operation.
     */
    public void setProperties(ArrayList properties) {
        OperationProperty p; 
        
        for (int i = 0; i < properties.size(); i++) {
            p = (OperationProperty)properties.get(i);
            
            try {
                if (p.getName().equals("Maximum number of terms")) {
                    max = Integer.parseInt(p.getValue()); 
                } else if (p.getName().equals("Get by weight")) {
                    byWeight = Boolean.parseBoolean(p.getValue());
                } else if (p.getName().equals("Keep words with digits")) {
                    digits = Boolean.parseBoolean(p.getValue());
                } else if (p.getName().equals("Keep the word 'more'")) {
                    more = Boolean.parseBoolean(p.getValue());
                } else if (p.getName().equals("Keep the word 'read'")) {
                    read = Boolean.parseBoolean(p.getValue());
                }  
            } catch (Exception e) {
                System.out.println("Error parsing property " + p.getName());
            }
        }
    }
}
