package datapreparation.partofspeech;

import datapreparation.TextOperation;
import datamodel.DataSet;
import datamodel.Text;
import java.util.ArrayList;
import java.io.File;
import java.text.BreakIterator;

import edu.stanford.nlp.tagger.maxent.MaxentTagger;
import control.Globals;
import gui.MainFrame;


/**
 * Class implementing the part-of speech operation. It uses a pre-trained Maximum Entropy
 * tagger to tag a sentence, all words are extracted, and the tagged text is stored.
 *
 * @author Ole Kristian Fivelstad
 */
public class PartOfSpeech extends TextOperation {
    private String stringText;
    private MaxentTagger tagger=null;   
    private ArrayList<String> dividers;
    private MainFrame mainFrame;



    /**
     * Constructor for the part-of-speech operation. Initializes a vector of characters
     * that divide the text into sentences or part of sentences. For example,
     * , ) ( and so on.
     */
    public PartOfSpeech(){

        dividers = new ArrayList<String>();
        dividers.add("'");
        dividers.add("");
        dividers.add("(");
        dividers.add("?");
        dividers.add(".");
        dividers.add(",");
        dividers.add("!");
        dividers.add(")");
        dividers.add("");
        dividers.add(":");
        dividers.add(";");

    }

    /**
     * Method for performing the operation.
     */
    public void performOperation(DataSet dataSet) {
        mainFrame = MainFrame.getInstance();
        try {
            //Loading tagger data
            tagger = new MaxentTagger("tagger"+File.separator+"train-wsj-0-18.holder");
        } catch (Exception e) {
            System.out.println("Tagger data not available...");
        }
        ArrayList<Text> texts = dataSet.getTexts();
        int numberOfTexts = texts.size();
        if(tagger!=null){
            for (int i = 0; i < numberOfTexts; i++) {
                texts.add(i, tagTerms((Text)texts.remove(i)));
                if(Globals.showGui){
                    int progress = (int)((double)(i + 1)/(double)numberOfTexts * 100);
                    mainFrame.updateOperationProgress(progress);
                }
            }
        }
        dataSet.setPartOfSpeechPerformed(true);
    }

    /**
     * Method for tagging a text
     *
     * @param text
     * @return newText
     */
    public Text tagTerms(Text text){
        stringText = text.toString();
        Text newText = new Text(text.getTid(), stringText);
        try {
            //tagging news item with part-of-speech tags
            stringText = stringText.replaceAll("per cent", "percent");
            stringText = tagger.tagString(stringText);
            newText.changeText(stringText);
        }catch (Exception e) {
            e.printStackTrace();
        }               
        ArrayList sentences = findSentences(stringText);
        for(int i=0;i<sentences.size();i++){
            String sentence = sentences.get(i).toString();
            extractTermsFromSentence(sentence, newText);
        }
        return newText;
    }

    /**
     * Method for extracting all the terms from a sentence
     *
     * @param sentence
     * @param newText
     */
    private void extractTermsFromSentence(String sentence, Text newText) {

        String words[] = sentence.split(" ");
        String word="";

        for(int i=0; i<words.length;i++){
            word = words[i];
            if(!word.trim().equals("")){
                if(word.indexOf("/CD")==-1){
                    newText.addTerm(removeChars(word), 1, 0);
                } else {
                    newText.addTerm(word, 1, 0);
                }
            }
        }
    }

    /**
     * Method for finding the sentences in a text.
     * @param text The text
     * @return ArrayList containing the sentences
     */
    public ArrayList findSentences(String text){
        ArrayList<String> sentences = new ArrayList<String>();
        BreakIterator bIterator = BreakIterator.getSentenceInstance();
        bIterator.setText(text);
        int start = bIterator.first();
        for (int end = bIterator.next();
             end != BreakIterator.DONE;
             start = end, end = bIterator.next()) {
            int startOfTag = text.indexOf("/", end);
            if(startOfTag!=-1){
                int endOfTag = text.indexOf(" ", startOfTag);
                end=endOfTag;
            }
            if(start>=0&&end>start){
                sentences.add(text.substring(start,end).trim());
            }
        }
        return sentences;
    }   

    /**
     * Method for removing characters from a word
     * @param word The word to clean
     * @return word after characters are removed
     */
    public String removeChars(String word){
        word=word.replaceAll("\\.","");
        word=word.replaceAll("\\,","");
        word=word.replaceAll("\\'","");
        word=word.replaceAll("\\!","");
        word=word.replaceAll("\\?","");
        word=word.replaceAll("\\(","");
        word=word.replaceAll("\\)","");
        word=word.replaceAll("\\;","");
        word=word.replaceAll("\\:","");
        return word;
    }



    /**
     * Method for getting the porperties.
     */
    public ArrayList getProperties() {
        return null;
    }

    /**
     * Method for setting the properties.
     */
    public void setProperties(ArrayList properties) {
    }

}
