package datapreparation.weighting;

import java.util.ArrayList;
import java.util.Hashtable;

import datamodel.DataSet;
import datamodel.Term;
import datamodel.Text;

/**
 * Class implementing a term collection. A term collection consists of a list
 * of terms with total frequence (sum of frequence in all texts) and the number
 * of text the term is present in.
 * 
 * @author Kjell-Inge Skogstad and Trond ivind Eriksen
 */
public class TermCollection {
	
	private Hashtable<String, TermDetails> hashtable;
	
	/**
	 * Constructor creating a new hashtable for storing terms.
	 */
	public TermCollection() {
		hashtable = new Hashtable<String, TermDetails>();
	}
	
	/**
	 * Method for filling the collection. A dataSet is given and parsed so that
	 * all individual terms are stored in the collection.
	 * 
	 * @param set The dataset
	 */
	public void fillCollection(DataSet set) {
		for (int i = 0; i < set.getNofTexts(); i++) {
			handleText((Text)set.getTexts().get(i));
		}
	}
	
	/**
	 * Method for dealing with all terms in a given text. Inserting the terms 
	 * into the collection.
	 * 
	 * @param text The text
	 */
	private void handleText(Text text) {

        ArrayList terms = text.getTerms();
		Term term;
		TermDetails td;
		
		for (int i = 0; i < terms.size(); i++) {
			term = (Term)terms.get(i);
			
			//Test if term is already present in the collection
			if (hashtable.containsKey(term.toString().toLowerCase())) {
				td = hashtable.get(term.toString().toLowerCase());
				td.addToFreq(term.getFrequency());
			} else {
				td = new TermDetails();
				td.addToFreq(term.getFrequency());
				hashtable.put(term.toString().toLowerCase(), td);
			}
		}
	}
	
	/**
	 * Method for getting the total frequency of a given term.
	 * 
	 * @param term The term
	 * @return The frequency
	 */
	public int getFrequency(String term) {
		return hashtable.get(term.toLowerCase()).freq;
	}
	
	/**
	 * Method for getting the number of texts the given term is present in.
	 * 
	 * @param term The term
	 * @return The number of texts
	 */
	public int getNofTexts(String term) {
		return hashtable.get(term.toLowerCase()).nofTexts;
	}
	
	/**
	 * Internal class for representing the details, frequency and number
	 * of texts, of a term.
	 * 
	 * @author Kjell-Inge Skogstad and Trond ivind Eriksen
	 */
	private class TermDetails {
		
		private int freq;
		private int nofTexts;
		/**
		 * Constructor creating a new detail container.
		 */
		public TermDetails() {
			freq = 0;
			nofTexts = 0;
		}
		
		/**
		 * Method for adding to the frequency. Also updates the number of texts
		 * the term is present in.
		 * 
		 * @param add Frequency to add
		 */
		public void addToFreq(int add) {
			freq += add;
			nofTexts++;
		}
    }
}
