package datapreparation.wordsensedisambiguation;

import datapreparation.TextOperation;
import datapreparation.stopword.EnglishStopwords;
import datamodel.DataSet;
import datamodel.Text;
import datamodel.Term;

import java.util.ArrayList;
import java.util.StringTokenizer;
import net.didion.jwnl.JWNLException;
import net.didion.jwnl.data.POS;
import net.didion.jwnl.data.IndexWord;
import net.didion.jwnl.data.Synset;
import net.didion.jwnl.data.PointerUtils;
import net.didion.jwnl.data.list.PointerTargetNodeList;
import net.didion.jwnl.data.list.PointerTargetNode;
import net.didion.jwnl.dictionary.Dictionary;
import control.OperationProperty;
import control.Globals;
import gui.MainFrame;

/**
 * Class implementing Word Sense Disambiguation using the Lesk and adapted Lesk algorithms.
 * The operation takes as input a tagged text, and a list of terms. The operation then tries
 * to disambiguate the terms tagged as nouns using the their context in the text.
 *
 *
 * @author Ole Kristian Fivelstad
 *
 */
public class WordSenseDisambiguation  extends TextOperation {

    private String stringText;
    private ArrayList<String> contextCandidates;
    private ArrayList<String> disambiguatedWords;
    private int prevPos=0;
    private int candidates=0;
    private int disambiguated=0;
    private int nonAmbigousNouns=0;
    private EnglishStopwords stopwords;

    //Some parameters for the operation
    private boolean adaptedLesk = true;
    private int context=3;
    private boolean keepAdjectives=false;
    private boolean keepVerbs=false;


    /**
     * Method for performing the operation.
     *
     * @param dataSet The dataset being used.
     */
    public void performOperation(DataSet dataSet) {
        if(dataSet.isCollocationPerformed()||dataSet.isPartOfSpeechPerformed()){
            candidates=0;
            disambiguated=0;
            nonAmbigousNouns=0;
            MainFrame mainFrame = MainFrame.getInstance();
            ArrayList<Text> texts = dataSet.getTexts();
            stopwords=new EnglishStopwords();
            int numberOfTexts = texts.size();
            for (int i = 0; i < numberOfTexts; i++) {
                prevPos=0;
                texts.add(i, disambiguateTermsInText((Text)texts.remove(i)));
                if(Globals.showGui){
                    int progress = (int)((double)(i + 1)/(double)numberOfTexts * 100);
                    mainFrame.updateOperationProgress(progress);
                }
            }

            //Displaying some statistics regarding the WSD-process
            double percentage = ((double)disambiguated)/candidates;
            System.out.println("Candidates in total: "+candidates);
            System.out.println("Candidates disambiguated: "+disambiguated);
            System.out.println("Percentage of candidates disambiguated: "+(percentage*100)+"%");
            System.out.println("In addition, "+nonAmbigousNouns+" nouns only had one sense.");
            System.out.println("In total: "+(nonAmbigousNouns+disambiguated)+" disambiguated nouns");
            dataSet.setWsdPerformed(true);
        } else {
            System.out.println("POS or collocation must be performed before WSD...");
        }
    }

    /**
     * The context is being initialized. This is done by tokenizing the full text, and
     * adding all terms tagged as nouns, adjectives, verbs to a list. Collocations are
     * also added.
     */
    private void initializeContext(){
        contextCandidates=new ArrayList<String>();
        String text = stringText;
        text=removeChars(text);
        StringTokenizer st = new StringTokenizer(stringText);
        while(st.hasMoreTokens()){
            String candidate = st.nextToken();
            if((candidate.contains("/NN")||candidate.contains("/JJ")||candidate.contains("/VB"))
                    &&(!candidate.contains("/NNP"))){
                contextCandidates.add(candidate);
            } else if( candidate.contains("/COLL")) {
                //Collocations are added if they are present in WordNet.
                //Otherwise it is split up, and each of the two nouns are added.
                String temp = candidate.substring(0, candidate.indexOf("/COLL"));
                try {
                    IndexWord collCandidate = Dictionary.getInstance().getIndexWord(POS.NOUN, temp);
                    if(collCandidate!=null){
                        contextCandidates.add(candidate);
                    } else {
                        String tempOne = temp.substring(0, temp.indexOf("_"));
                        String tempTwo = temp.substring(temp.indexOf("_")+1, temp.length());
                        tempOne +="/NN";
                        tempTwo +="/NN";
                        contextCandidates.add(tempOne);
                        contextCandidates.add(tempTwo);
                    }
                } catch (JWNLException e) {
                    e.printStackTrace();
                }
            }
        }
    }

    /**
     * Method for getting the surrounding context of a word.
     * @param word The word.
     * @return A vector containing the context.
     */
    private Object[] getContext(String word){
        word=removeChars(word);
        Object[] contextWords;
        ArrayList<String> temporary = new ArrayList<String>();
        ArrayList<String> tempContext = (ArrayList<String>)contextCandidates.clone();
        int counter=0;
        if(tempContext.size()<=((context*2)+1)){
            //Number of possible candidates is less than wanted context
            //all words are therefore added.
            contextWords=new Object[contextCandidates.size()];
            tempContext.remove(word);
            contextWords=tempContext.toArray();
            return contextWords;
        } else{
            //Finding the context of a word by getting one and one word to both
            //the left and right of the word.
            int position=findPosition(tempContext.toArray(), prevPos, word);
            prevPos=position;
            int pos=0;
            while(counter<(context*2)){
                pos++;
                int left=position-pos;
                int right=position+pos;
                if(left<0 && right>(tempContext.size()-1)){
                    break;
                }
                if(left>-1){
                    String temp = tempContext.get((left));
                    if(!temporary.contains(temp)&&!word.equals(temp)){

                        temporary.add(temp);
                        counter++;
                        if(counter==context*2) break;
                    }
                }
                if(right<tempContext.size()){
                    String temp = tempContext.get((right));
                    if(!temporary.contains(temp)&&!word.equals(temp)){
                        temporary.add(temp);
                        counter++;
                        if(counter==context*2) break;
                    }
                }

            }
        }
        contextWords=temporary.toArray();
        return contextWords;
    }

    /**
     * Method for performing the disambiguation of the terms in a text.
     * @param text The text
     * @return The new text
     */
    public Text disambiguateTermsInText(Text text){
        stringText=text.toString();
        Text newText = new Text(text.getTid(), stringText);
        ArrayList terms = text.getTerms();
        disambiguatedWords=new ArrayList<String>();
        initializeContext();

        for(int i=0;i<terms.size();i++){
            Term temp = ((Term)terms.get(i));
            String term = temp.toString();

            if(term.contains("/COLL")) {
                //collocations are added to the new text
                String collString = term.substring(0, term.indexOf("/COLL"));
                try{
                    IndexWord candidate = Dictionary.getInstance()
                            .getIndexWord(POS.NOUN, collString.toLowerCase());
                    if(candidate==null){
                        newText.addTerm(term, temp.getFrequency(), temp.getWeight());
                    } else{
                        newText.addTerm(collString+"#1", temp.getFrequency(), temp.getWeight());
                    }
                } catch(Exception e){
                    newText.addTerm(term, temp.getFrequency(), temp.getWeight());
                }
            } else if(term.contains("NNP")){
                //proper nouns are added to the new text.
                newText.addTerm(term, temp.getFrequency(), temp.getWeight());
            } else if(term.contains("/NN")){
                //nouns are disambiguated
                Object[] context = getContext(term);
                term=term.substring(0, term.indexOf("/"));
                try {
                    IndexWord candidate = Dictionary.getInstance().getMorphologicalProcessor()
                            .lookupBaseForm(POS.NOUN, term.toLowerCase());
                    if(candidate==null){
                        newText.addTerm(term, temp.getFrequency(), 0);
                    } else{
                            if(candidate.getSenseCount()>1){
                                candidates++;
                                int result = -1;
                                if(!adaptedLesk){
                                    result=originalLeskAlgorithm(candidate, context);
                                } else{
                                    result=adaptedLeskAlgorithm(candidate, context);
                                }
                                if(result!=-1){
                                    disambiguated++;
                                    newText.addTerm(candidate.getLemma()+"#"+result, temp.getFrequency(), 0);
                                    disambiguatedWords.add(candidate.getLemma()+"#"+result);
                                } else{
                                    newText.addTerm(candidate.getLemma()+"/NN", temp.getFrequency(), 0);
                                }

                            } else{
                                nonAmbigousNouns++;
                                newText.addTerm(candidate.getLemma()+"#1", temp.getFrequency(), 0);
                                disambiguatedWords.add(candidate.getLemma()+"#1");
                            }
                    }

                } catch (JWNLException e) {
                    e.printStackTrace();
                }

            //Various other terms may be addded to the new text.
            } else if(term.contains("/JJ")&&keepAdjectives){
                newText.addTerm(term, temp.getFrequency(), temp.getWeight());
            } else if(term.contains("/VB")&&keepVerbs){
                newText.addTerm(term, temp.getFrequency(), temp.getWeight());
            } else if(term.contains("/RB")||term.contains("/CD")) {
                newText.addTerm(term, temp.getFrequency(), temp.getWeight());
            }
        }
        return newText;
    }


    /**
     * Method for finding the position of a term in a vector.
     * @param array The vector
     * @param fromIndex Where to start looking
     * @param object The term
     * @return The position of the term
     */
    private int findPosition(Object[] array, int fromIndex, String object){
        if(fromIndex>=array.length){
            return -1;
        }else{
            int position = 0;
            for(int i=fromIndex;i<array.length;i++){
                if(((String)array[i]).equals(object)){
                    position=i;
                    break;
                }
            }
            return position;
        }
    }

    /**
     * Method for finding out if a word has been disambiguated.
     *
     * @param word The word
     * @return The result
     */
    private boolean termIsDisambiguated(String word){
        String temp;
        for(int i=0;i<disambiguatedWords.size();i++){
            temp=disambiguatedWords.get(i);
            temp=temp.substring(0,temp.indexOf("#"));
            if(temp.equals(word)){
                return true;
            }
        }
        return false;
    }

    /**
     * Method for getting the sense of an already disambiguated word
     * @param word The word
     * @return The sense
     */
    private int getSense(String word){
        int sense=-1;
        for(int i=0;i<disambiguatedWords.size();i++){
            String temp=disambiguatedWords.get(i);
            String temp2=temp.substring(0,temp.indexOf("#"));
            if(temp2.equals(word)){
                String number = temp.substring((temp.indexOf("#")+1), temp.length());
                return Integer.parseInt(number);
            }
        }
        return sense;
    }

    /**
     * Method implementing the adapted Lesk algorithm.
     *
     * @param candidate The word being disambiguated
     * @param contextArray The context of the word
     * @return The sense, -1 if no conclusion can be made
     */
    private int adaptedLeskAlgorithm(IndexWord candidate, Object[] contextArray){

        int result=-1;
        //vector holding the score of each sense
        int[] sense=new int[candidate.getSenseCount()];


        PointerTargetNodeList candidateHypernyms;
        PointerTargetNodeList contextHypernyms;

        PointerTargetNodeList candidateHyponyms;
        PointerTargetNodeList contextHyponyms;

        PointerTargetNodeList candidateMeronyms;
        PointerTargetNodeList contextMeronyms;

        PointerTargetNodeList candidateHolonyms;
        PointerTargetNodeList contextHolonyms;

        PointerTargetNodeList contextSynonyms;

        sense=new int[candidate.getSenseCount()];

        for(int i=0;i<candidate.getSenseCount();i++){
            try {

                Synset candidateSense = candidate.getSense(i+1);
                String candidateGloss = candidateSense.getGloss();
                candidateGloss=removeChars(candidateGloss);
                candidateGloss=stopwords.removeStopwords(candidateGloss.trim());

                candidateHypernyms = PointerUtils.getInstance().getDirectHypernyms(candidateSense);
                candidateHyponyms  = PointerUtils.getInstance().getDirectHyponyms(candidateSense);
                candidateMeronyms = PointerUtils.getInstance().getDirectHypernyms(candidateSense);
                candidateHolonyms  = PointerUtils.getInstance().getDirectHyponyms(candidateSense);


                String candidateHypernymGloss = concatenateGlosses(candidateHypernyms);
                String candidateHyponymGloss = concatenateGlosses(candidateHyponyms);
                String candidateMeronymGloss = concatenateGlosses(candidateMeronyms);
                String candidateHolonymGloss = concatenateGlosses(candidateHolonyms);

                String candidateGlosses = candidateGloss+" "+candidateHypernymGloss+" "+candidateHyponymGloss+
                        " "+candidateHolonymGloss+" "+candidateMeronymGloss;

                for(int n=0;n<contextArray.length;n++){
                    String contextTemp = "";
                    contextTemp = (String)contextArray[n];
                    POS candPos = POS.NOUN;
                    //finding out the word class of the context word.
                    if(contextTemp.contains("/VB")){
                        candPos = POS.VERB;
                    } else if(contextTemp.contains("/JJ")){
                        candPos = POS.ADJECTIVE;
                    }
                    contextTemp=contextTemp.substring(0, contextTemp.indexOf("/"));
                    IndexWord context =  Dictionary.getInstance().getMorphologicalProcessor()
                            .lookupBaseForm(candPos, contextTemp.toLowerCase());
                    if(context!=null){
                        int start=1;
                        int end=context.getSenseCount();
                        if(candPos==POS.NOUN && termIsDisambiguated(context.getLemma())){
                            start=getSense(context.getLemma());
                            end=start;
                        }
                        for(int j=start;j<=end;j++){
                            Synset contextSense = context.getSense(j);
                            String contextGloss = contextSense.getGloss();
                            contextGloss = removeChars(contextGloss);
                            contextGloss = stopwords.removeStopwords(contextGloss.trim());

                            if(context.getPOS()==POS.NOUN){
                                //System.out.println("Context word is noun");
                                contextHypernyms = PointerUtils.getInstance().getDirectHypernyms(contextSense);
                                contextHyponyms = PointerUtils.getInstance().getDirectHyponyms(contextSense);
                                contextMeronyms = PointerUtils.getInstance().getMeronyms(contextSense);
                                contextHolonyms = PointerUtils.getInstance().getHolonyms(contextSense);

                                String contextHypernymGloss = concatenateGlosses(contextHypernyms);
                                String contextHyponymGloss = concatenateGlosses(contextHyponyms);
                                String contextMeronymGloss = concatenateGlosses(contextMeronyms);
                                String contextHolonymGloss = concatenateGlosses(contextHolonyms);

                                String contextGlosses = contextGloss+" "+contextHypernymGloss+" "
                                        +contextHyponymGloss+" "+contextMeronymGloss+" "+contextHolonymGloss;
                                sense[i]+=findNumberOfMatches(candidateGlosses, contextGlosses);


                            } else if(context.getPOS()==POS.VERB){
                                contextHypernyms = PointerUtils.getInstance().getDirectHypernyms(contextSense);
                                contextHyponyms = PointerUtils.getInstance().getDirectHyponyms(contextSense);

                                String contextHypernymGloss = concatenateGlosses(contextHypernyms);
                                String contextHyponymGloss = concatenateGlosses(contextHyponyms);

                                String contextGlosses = contextGloss+" "+contextHypernymGloss+" "
                                        +contextHyponymGloss;

                                //calculating the number of matches
                                sense[i]+=findNumberOfMatches(candidateGlosses,contextGlosses);

                            } else{
                                contextSynonyms = PointerUtils.getInstance().getSynonyms(contextSense);
                                String synonymGloss = concatenateGlosses(contextSynonyms);
                                String contextGlosses = contextGloss + " " +synonymGloss;
                                sense[i]+=findNumberOfMatches(candidateGlosses,contextGlosses);//+


                            }
                        }
                    }
                }
            } catch (JWNLException e) {
                e.printStackTrace();
            }
        }

        //Finding out which sense got the highest score.
        int highest=0;
        for(int p=0;p<sense.length;p++){
            if(sense[p]>highest){
                highest=sense[p];
                result=p+1;
            }
        }
        return result;
    }

    /**
     * Method for finding the number of mathces between two strings.
     *
     * @param candidate The first string
     * @param context The second string
     * @return The number of matches
     */
    private int findNumberOfMatches(String candidate, String context){
        int result=0;
        StringTokenizer stCandidate = new StringTokenizer(candidate);
        StringTokenizer stContext = new StringTokenizer(context);


        String[] candWords = new String[stCandidate.countTokens()];
        String[] contWords = new String[stContext.countTokens()];

        int counter=0;
        while(stCandidate.hasMoreTokens()){
            candWords[counter]=stCandidate.nextToken();
            counter++;
        }
        counter=0;
        while(stContext.hasMoreTokens()){
            contWords[counter]=stContext.nextToken();
            counter++;
        }

        for(int i=0; i<candWords.length; i++){
            String tempCand = candWords[i];
            for(int j=0; j<contWords.length; j++){
                String tempCont = contWords[j];
                if(tempCand.equals(tempCont)){
                    result++;
                }
            }
        }
        return result;
    }

    /**
     * Method for concatenating the glosses of a list of related words.
     *
     * @param ptnList The list of the words
     * @return The combined gloss
     */
    private String concatenateGlosses(PointerTargetNodeList ptnList){
        String result="";

        for(int i=0;i<ptnList.size();i++){
            PointerTargetNode ptn = (PointerTargetNode)ptnList.get(i);
            String gloss = ptn.getSynset().getGloss();
            gloss=removeChars(gloss);
            gloss=stopwords.removeStopwords(gloss.trim());
            result+=gloss+" ";
        }

        return result;
    }

    /**
     * Method for removing a set of characters from a word.
     * @param word The word
     * @return The new word
     */
    public String removeChars(String word){
        word=word.replaceAll("\\.","");
        word=word.replaceAll("\\:","");
        word=word.replaceAll("\\;","");
        word=word.replaceAll("\\,","");
        word=word.replaceAll("\\'","");
        word=word.replaceAll("\\!","");
        word=word.replaceAll("\\?","");
        word=word.replaceAll("\\(","");
        word=word.replaceAll("\\)","");
        word=word.replaceAll("\"","");
        return word;
    }

    /**
     * Method implementing the original Lesk algorithm.
     *
     * @param candidate The word being disambiguated
     * @param contextWords The context of the word
     * @return The most likely sense
     */
    private int originalLeskAlgorithm(IndexWord candidate, Object[] contextWords){
        int result=-1;
        //Vector holding the score of each sense
        int[] senses=new int[candidate.getSenseCount()];
        String term="";
        for(int g=0; g<contextWords.length;g++){
            try{
                term = contextWords[g].toString();
                POS candPos = POS.NOUN;
                if(term.contains("/VB")){
                    candPos = POS.VERB;
                } else if(term.contains("/JJ")){
                    candPos = POS.ADJECTIVE;
                }
                String termPos = term;
                term=term.substring(0, term.indexOf("/"));
                IndexWord contextWord = Dictionary.getInstance().getMorphologicalProcessor()
                        .lookupBaseForm(candPos, term.toLowerCase());
                if(contextWord!=null){
                    Synset[] candidateSet = candidate.getSenses();
                    //Synset[] contextSet = contextWord.getSenses();
                    int start=1;
                    int end=contextWord.getSenseCount();
                    if(candPos==POS.NOUN&&termIsDisambiguated(contextWord.getLemma())){
                        start=getSense(contextWord.getLemma());
                        end=start;
                    }

                    for(int i=0; i<candidateSet.length;i++){
                        String candidateGloss=candidateSet[i].getGloss();
                        candidateGloss=removeChars(candidateGloss);
                        candidateGloss=stopwords.removeStopwords(candidateGloss).trim();
                        if(!candidateGloss.contains(candidate.getLemma())){
                            candidateGloss=candidateGloss+" "+candidate.getLemma();
                        }
                        StringTokenizer candTokenizer = new StringTokenizer(candidateGloss);
                        String[] candArray = new String[candTokenizer.countTokens()];
                        int counter=0;
                        while(candTokenizer.hasMoreTokens()){
                            candArray[counter]=candTokenizer.nextToken();
                            counter++;
                        }
                        for(int j=start; j<=end; j++){
                            String contextGloss=contextWord.getSense(j).getGloss();
                            contextGloss=removeChars(contextGloss);
                            contextGloss=stopwords.removeStopwords(contextGloss).trim();
                            if(!contextGloss.contains(contextWord.getLemma())){
                                contextGloss=contextGloss+" "+contextWord.getLemma();
                            }
                            StringTokenizer contTokenizer = new StringTokenizer(contextGloss);
                            String[] contextArray = new String[contTokenizer.countTokens()];
                            counter=0;
                            while(contTokenizer.hasMoreTokens()){
                                contextArray[counter]=contTokenizer.nextToken();
                                counter++;
                            }
                            for(int k=0;k<candArray.length;k++){
                                String tempCand=candArray[k];
                                for(int l=0;l<contextArray.length;l++){
                                    String tempContext=contextArray[l];
                                    if((tempCand.equals(tempContext))){
                                        senses[i]+=1;
                                    }
                                }
                            }


                        }
                    }
                }
            } catch (Exception e){
                e.printStackTrace();
            }
        }

        //Finding the sense with the highest score
        int highest = 0;
        for(int k=0; k<senses.length;k++){
            if(senses[k]>highest){
                highest=senses[k];
                result=k+1;
            }
        }
        return result;

    }



    /**
     * Method for getting the properties
     *
     * @return properties
     */
    public ArrayList getProperties() {
        ArrayList properties = new ArrayList();

        properties.add(new OperationProperty("Use adapted Lesk", adaptedLesk+""));
        properties.add(new OperationProperty("Size of context",context+""));
        properties.add(new OperationProperty("Keep verbs", keepVerbs+""));
        properties.add(new OperationProperty("Keep adjectives", keepAdjectives+""));

        return properties;
    }

    /**
     * Method for setting the properties
     *
     * @param properties
     */
    public void setProperties(ArrayList properties) {
        OperationProperty p;

        for (int i = 0; i < properties.size(); i++) {
            p = (OperationProperty)properties.get(i);
            try {
                if (p.getName().equals("Use adapted Lesk")) {
                    adaptedLesk = Boolean.parseBoolean(p.getValue());
                } else if (p.getName().equals("Size of context")) {
                    context = Integer.parseInt(p.getValue());
                } else if (p.getName().equals("Keep verbs")) {
                    keepVerbs = Boolean.parseBoolean(p.getValue());
                } else if (p.getName().equals("Keep adjectives")) {
                    keepAdjectives = Boolean.parseBoolean(p.getValue());
                }
            } catch (Exception e) {
                System.out.println("Error parsing property " + p.getName());
            }
        }
    }
}
