package gui;

import gui.panel.*;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Container;
import java.awt.Dimension;
import java.util.ArrayList;

import javax.swing.*;

import control.Control;

/**
 * The main gui class. This class implements the graphical user interface of 
 * the application. 
 * 
 * @author Kjell-Inge Skogstad, Trond ivind Eriksen and Ole Kristian Fivelstad
 */
public class MainFrame extends JFrame {

    /* MainFrame instance */
    private static MainFrame INSTANCE;
    
    private Control control;
    private RunPanel runPanel;
    private ProcessPanel processPanel;
    private DocReportPanel docReportPanel;
    private DocPanel docPanel;
    private SourcePanel sourcePanel;
    private OperationProgressPanel operationProgressPanel;

  
    /**
     * Constructor initializing the graphical user interface.
     * 
     * @param control The control instance
     */
    public MainFrame(Control control) {
        super("TTM Testbench");
        //JFrame frame = new JFrame("TTM Testbench");

        INSTANCE = this;
        
        this.control = control;

        addComponents();
        setJMenuBar(new Menu(control));
        setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
        pack();
        setLocationRelativeTo(null);
        setVisible(true);
    }

    
    /**
     * Method returning the instance of the MainFrame.
     * 
     * @return The MainFrame instance
     */
    public static MainFrame getInstance() {
        return INSTANCE;
    }
    /**
     * Method for getting the Control instance.
     * 
     * @return The Control instance
     */
    public Control getControl() {
        return control;
    }


    
    /**
     * Method that adds all components to the gui.
     */

    private void addComponents() {
        Container pane = getContentPane();
        JSplitPane splitPane = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT);
        JPanel menuPanel = new JPanel(new BorderLayout());
        JPanel menuPanel2 = new JPanel(new BorderLayout());
        JPanel menuPanelHelp = new JPanel(new BorderLayout());
        JPanel feedbackPanel = new JPanel(new BorderLayout());

        sourcePanel = new SourcePanel(control);

        pane.setLayout(new BorderLayout());
        
        /* Initializes the left process panel */    

        runPanel = new RunPanel(control);
        processPanel = new ProcessPanel(control);
        operationProgressPanel = new OperationProgressPanel(control);
        
        menuPanel.add(sourcePanel, BorderLayout.NORTH);
        menuPanel.add(processPanel, BorderLayout.CENTER);
        menuPanel.add(runPanel, BorderLayout.SOUTH);
        menuPanel2.add(operationProgressPanel, BorderLayout.PAGE_START);

        menuPanelHelp.setPreferredSize(new Dimension(250, 700));
        menuPanelHelp.add(menuPanel, BorderLayout.NORTH);
        menuPanelHelp.add(menuPanel2, BorderLayout.CENTER);

        splitPane.setLeftComponent(menuPanelHelp);

        /* Initializes the right feedback panel */
        docReportPanel = new DocReportPanel();
        docPanel = new DocPanel(this);
        
        feedbackPanel.add(docReportPanel, BorderLayout.CENTER);
        feedbackPanel.add(docPanel, BorderLayout.SOUTH);
        
        feedbackPanel.setPreferredSize(new Dimension(600, 700));
        feedbackPanel.setBackground(Color.red);
        splitPane.setRightComponent(feedbackPanel);
        
        pane.add(splitPane, BorderLayout.CENTER);
    }
    
    /**
     * Method used to retrieve the process created. The process
     * consists of a list of operations.
     * 
     * @return The process
     */
    public ArrayList getProcess() {
        return processPanel.getProcess();
    }
    
    /**
     * Method used to update the progress bar. Progress is a number
     * between 0 and 100.
     * 
     * @param newProgress The new progress
     */
    public void updateProgress(int newProgress) {
        runPanel.updateProgress(newProgress);
    }
    /**
     * Method used to update the operation progress bar. Progress is a number
     * between 0 and 100.
     *
     * @param newProgress The new progress
     */
    public void updateOperationProgress(int newProgress) {
        operationProgressPanel.updateProgress(newProgress);
    }

    public void invalidOperationProgress(){
        operationProgressPanel.invalidProgress();
    }
    
    /**
     * Method used to specify which document report is shown in gui given by
     * the document index.
     * 
     * @param index The index
     */
    public void setDocIndex(int index) {
        docReportPanel.setPrintedReport(control.getDocReport(index));
    }
    
    /**
     * Method used to update the sliderposition.
     * 
     * @param index The new position
     */
    public void setSliderIndex(int index) {
        docPanel.setSlider(index);
    }
    
    /**
     * Method used to configurate the gui by setting the number of documetns
     * to be shown.
     * 
     * @param max The number of documents
     */
    public void configGui(int max, String name, String language, String nofTexts) {
        docPanel.configSlider(max);
        sourcePanel.setName(name);
        sourcePanel.setLanguage(language);
        sourcePanel.setNofTexts(nofTexts);
    }
    
    /**
     * Method used to print messages in a popup window.
     * 
     * @param string The message
     */
    public void printMessage(String string) {
       JOptionPane.showMessageDialog(null,
                string, 
                "Warning", 
                JOptionPane.WARNING_MESSAGE);
    }
}
