package gui.dialog;

import gui.MainFrame;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;

import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.event.TableModelListener;
import javax.swing.table.TableModel;

import control.Operation;
import control.OperationProperty;

/**
 * Class implementing a window for editing the properties of the
 * operations used in this application. Here the user can select
 * and change the properties of a given operation.
 *
 * @author Kjell-Inge Skogstad and Trond ivind Eriksen
 */
public class PropertiesDialog extends JDialog {

    private JButton ok;
    private Operation op;
    private PropertiesModel model;

    /**
     * Constructor creating a new winodw and adding all components and
     * listeners.
     *
     * @param op The operation
     */
    public PropertiesDialog(Operation op) {
        super(MainFrame.getInstance(), "Properties", true);

        this.op = op;

        addComponents();
        addListeners();

        pack();
        setLocationRelativeTo(MainFrame.getInstance());
        setVisible(true);
    }

    /**
     * Method adding all components to the window.
     */
    private void addComponents() {
        JPanel buttons = new JPanel(new FlowLayout(FlowLayout.RIGHT));
        JScrollPane scroll = new JScrollPane(
                JScrollPane.VERTICAL_SCROLLBAR_ALWAYS,
                JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
        JTable table;

        ok = new JButton("OK");
        buttons.add(ok);

        model = new PropertiesModel(op.getProperties());
        table = new JTable(model);

        scroll.setViewportView(table);
        scroll.setPreferredSize(new Dimension(400, 400));

        this.getContentPane().setLayout(new BorderLayout());
        this.getContentPane().add(scroll, BorderLayout.CENTER);
        this.getContentPane().add(buttons, BorderLayout.SOUTH);
    }

    /**
     * Method for adding all listeners to the components in this window.
     */
    private void addListeners() {
        ok.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                storeProperties();
                dispose();
            }});
    }

    /**
     * Method storing the new properties.
     */
    private void storeProperties() {
        op.setProperties(model.getProperties());
    }

    /**
     * Internal class representing a tablemodel of the properties of
     * the operation. Used in the visualization of the properties in a table.
     *
     * @author Kjell-Inge Skogstad and Trond ivind Eriksen
     */
    private class PropertiesModel implements TableModel {

        private ArrayList properties;

        /**
         * Constructor creating a new properties model.
         *
         * @param properties The properties
         */
        public PropertiesModel(ArrayList properties) {
            this.properties = properties;
        }

        /**
         * Method for getting the properties.
         *
         * @return The properties
         */
        public ArrayList getProperties() {
            return properties;
        }

        /**
         * Method for getting the number of properties.
         */
        public int getRowCount() {
            return properties.size();
        }

        /**
         * Method for getting the number of columns in the model.
         */
        public int getColumnCount() {
            return 2;
        }

        /**
         * Method for getting the name of each column.
         */
        public String getColumnName(int arg0) {
           switch (arg0) {
                case 0:
                    return "Name";
                case 1:
                    return "Value";
            }
            return "";
        }

        /**
         * Method for getting the class of the values in a column.
         */
        public Class<?> getColumnClass(int c) {
            return getValueAt(0, c).getClass();
        }

        /**
         * Method returning wether a cell is editable or not.
         */
        public boolean isCellEditable(int row, int col) {
            switch (col) {
                case 0:
                    return false;
                case 1:
                    return true;
            }
            return false;
        }

        /**
         * Method for getting the value at a specific cell.
         */
        public Object getValueAt(int row, int col) {
            switch (col) {
                case 0:
                    return ((OperationProperty)properties.get(row)).getName();
                case 1:
                    return ((OperationProperty)properties.get(row)).getValue().toString();
            }
            return null;
        }

        /**
         * Method for setting the value at a specific cell.
         */
        public void setValueAt(Object value, int row, int col) {
            switch (col) {
                case 1:
                    ((OperationProperty)properties.get(row)).setValue((String)value);
            }
        }

        /**
         * Method for adding a tablemodel listener. Not used.
         */
        public void addTableModelListener(TableModelListener arg0) {
        }

        /**
         * Method for removing a tablemodel listener. Not used.
         */
        public void removeTableModelListener(TableModelListener arg0) {
        }
    }
}
