package gui.dialog;

import gui.*;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.util.Calendar;
import java.util.GregorianCalendar;

import javax.swing.*;

import dataloader.FileHandler;
import control.Globals;
import modelling.fiti.report.FitiReport;

/**
 * Class implementing a window to present reports from modelling operations.
 * A operation specific report view is added to this window presenting the 
 * operation result.
 * 
 * @author Kjell-Inge Skogstad and Trond ivind Eriksen
 */
public class ReportDialog extends JDialog {

    private JButton ok;
    private JButton save;
    private ReportView comp;

    private String comment;
    private boolean gui;

    /**
     * Constructor creating a new window showing a modelling report. A operation
     * specific component (report view) is added to the window.
     * 
     * @param title The title
     * @param comp The report view
     */
    public ReportDialog(String title, ReportView comp) {
        super(MainFrame.getInstance(), title, true);
        gui= Globals.showGui;
        this.comp = comp;
        comment = "";

        addComponents();
        addListeners();
        pack();
        setLocationRelativeTo(MainFrame.getInstance());
        setVisible(true);
    }

    /**
     * Metehod adding all components to the window.
     */
    private void addComponents() {
        JPanel buttons = new JPanel(new FlowLayout(FlowLayout.RIGHT));

        ok = new JButton("OK");
        save = new JButton("Save as...");

        buttons.add(save);
        buttons.add(ok);

        comp.setPreferredSize(new Dimension(600, 600));

        this.getContentPane().setLayout(new BorderLayout());
        this.getContentPane().add(comp, BorderLayout.CENTER);
        this.getContentPane().add(buttons, BorderLayout.SOUTH);
    }

    /**
     * Method adding listeners to the components of the window.
     */
    private void addListeners() {
        ok.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                dispose();
            }});
        save.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                JFileChooser chooser = new JFileChooser();
                int returnVal = chooser.showSaveDialog(null);

                if(returnVal == JFileChooser.APPROVE_OPTION) {
                    new CommentDialog();

                    if (!comment.equals("")) {
                        comment  = "\nComment: \n" + comment + "\n";
                    }

                    MainFrame.getInstance().getControl().saveReport(
                            chooser.getSelectedFile().getPath(),
                            comp.getPrintableCopy() + comment);
                }
            }});
    }

    /**
     * Internal class for implementing a comment dialog. A dialog is shown 
     * where the user may comment a given report which is stored.
     * 
     * @author Kjell-Inge Skogstad and Trond ivind Eriksen
     */
    private class CommentDialog extends JDialog {

        private JTextArea textArea;
        private JButton ok;

        /**
         * Constructor creating the dialog.
         */
        public CommentDialog() {
            super(MainFrame.getInstance(), "Comment", true);

            addComponents();
            addListeners();

            pack();
            setLocationRelativeTo(MainFrame.getInstance());
            setVisible(true);
        }

        /**
         * Method for adding all components to the dialog.
         */
        private void addComponents() {
            JPanel buttons = new JPanel(new FlowLayout(FlowLayout.RIGHT));
            JScrollPane scroll = new JScrollPane(
                    JScrollPane.VERTICAL_SCROLLBAR_ALWAYS,
                    JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);

            ok = new JButton("OK");
            buttons.add(ok);

            textArea = new JTextArea();
            textArea.setLineWrap(true);
            textArea.setWrapStyleWord(true);

            scroll.setPreferredSize(new Dimension(300, 300));
            scroll.setViewportView(textArea);

            getContentPane().setLayout(new BorderLayout());
            getContentPane().add(scroll, BorderLayout.CENTER);
            getContentPane().add(buttons, BorderLayout.SOUTH);
        }

        /**
         * Method for adding all listeners.
         */
        private void addListeners() {
            ok.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    comment = textArea.getText();
                    dispose();
                }});
        }
    }
}
