package gui.panel;

import java.util.ArrayList;

import javax.swing.event.TableModelListener;
import javax.swing.table.TableModel;

import control.Operation;
import control.OperationConfig;
import control.Globals;
import gui.MainFrame;

/**
 * Class implementing a model of the process. This class is used by the 
 * process panel to generate a temporal text mining process.
 * 
 * @author Kjell-Inge Skogstad and Trond ivind Eriksen
 */
public class ProcessModel implements TableModel {

    private ArrayList operations;
    private ArrayList log;
    private ArrayList possibleOperations;
    private OperationConfig oc;
    private int operationCounter;

    /**
     * Constructor that initializes cocnfiguration of operations and creates
     * new lists to hold process information.
     */
    public ProcessModel() {
        operations = new ArrayList();
        log = new ArrayList();
        oc = new OperationConfig();
        possibleOperations = oc.getOperationDescriptions();
        operationCounter = 0;
    }

    public OperationConfig getOpConfig(){
        return oc;
    }

    /**
     * Method to add an operation to the process. The index shows the order
     * of the operations.
     * 
     * @param index The operation index
     */
    public void addOperation(int index) {

        if(Globals.showGui){
            /* Test if possible operations exist */
            if (possibleOperations.size() > 0) {
                //Test if index is valid, if not append operation at the end of the process
                if (index < 0 || index >= getRowCount()) {
                    try {
                        operations.add((Operation)oc.getOperation(
                                (String)possibleOperations.get(operationCounter)).newInstance());
                    } catch (Exception e) {
                        e.printStackTrace();
                    }

                    log.add(false);
                } else {
                    try {
                        operations.add(index, (Operation)oc.getOperation(
                                (String)possibleOperations.get(operationCounter)).newInstance());
                    } catch (Exception e) {
                        e.printStackTrace();
                    }
                    log.add(index, false);
                }
                operationCounter = (operationCounter + 1) % possibleOperations.size();
            }
        }else{
            /* Test if possible operations exist */
            if (possibleOperations.size() > 0) {
                try {
                    operations.add(operationCounter,(Operation)oc.getOperation((String)possibleOperations.get(index)).newInstance());                    
                    System.out.println("Adding operation: "+((Operation)operations.get(operations.size()-1)).getClass().getSimpleName());
                } catch (Exception e) {
                    if(Globals.showGui){
                        MainFrame.getInstance().printMessage("No more operations available.");
                    }
                }
                log.add(operationCounter, false);
                operationCounter = (operationCounter + 1) % possibleOperations.size();
            }
        }       
    }

    /**
     * Method returning a specific operation given by its index.
     * 
     * @param index The index
     * @return The operation
     */
    public Operation getOperation(int index) {
        return (Operation)operations.get(index);
    }

    /**
     * Method for removing an operation from the process given by its index.
     * 
     * @param index The operation index
     */
    public void removeOperation(int index) {
        if (index >= 0 && index < getRowCount()) {
            operations.remove(index);
            log.remove(index);
        }
    }

    /**
     * Method for retrieving the description of the possible operations. This
     * method is used by the process panel to find alternative operations.
     * 
     * @return List of possible operations
     */
    public ArrayList getPossibleOperations() {
        return possibleOperations;
    }

    /**
     * Method for getting the chosen process. This process is given as an 
     * arraylist of operations.
     * 
     * @return The process
     */
    public ArrayList getProcess() {
        ArrayList process = new ArrayList();
        Operation op;

        for (int i = 0; i < operations.size(); i++) {
            op = (Operation)operations.get(i);
            op.setLogResult(((Boolean)log.get(i)).booleanValue());
            process.add(op);
        }
        return process;
    }

    /**
     * Method returning the number of rows.
     */
    public int getRowCount() {
        return operations.size();
    }

    /**
     * Method returning the number of columns.
     */
    public int getColumnCount() {
        return 2;
    }

    /**
     * Method returning name of column.
     */
    public String getColumnName(int arg0) {
        String name = "";

        if (arg0 == 0) {
            name = "Nr";
        } else if (arg0 == 1) {
            name = "Operation";
        } else {
            name = "Log";
        }
        return name;
    }

    /**
     * Method returning the class type of a column.
     */
    public Class getColumnClass(int c) {
        return getValueAt(0, c).getClass();
    }

    /**
     * Method returning wether the cell is editable or not.
     */
    public boolean isCellEditable(int row, int column) {
        boolean edit = false;

        if (column > 0) {
            edit = true;
        }
        return edit;
    }

    /**
     * Method returning the value of a cell.
     */
    public Object getValueAt(int row, int column) {
        Object value = null;

        if (column == 0) {
            value = row + 1;
        } else if (column == 1) {
            value = oc.getOperationDescription(((Operation)operations.get(row)).getClass());
        } else {
            value = ((Boolean)log.get(row)).booleanValue();
        }
        return value;
    }

    /**
     * Method setting the value of a cell.
     */
    public void setValueAt(Object value, int row, int column) {
        if (row < operations.size()) {
            if (column == 1) {
                try {
                    Operation op = (Operation)oc.getOperation((String)value).newInstance();
                    Operation current = (Operation)operations.get(row);

                    if (oc.isEqual(op, current)) {
                        return;
                    }

                    operations.remove(row);
                    operations.add(row, op);
                } catch (Exception e) {
                    e.printStackTrace();
                }
            } else if (column == 2){
                log.remove(row);
                log.add(row, value);
            }
        }
    }

    /**
     * Method for adding a table model listener. Not used.
     */
    public void addTableModelListener(TableModelListener arg0) {
    }

    /**
     * Method for removing a table model listener. Not used.
     */
    public void removeTableModelListener(TableModelListener arg0) {
    }
}
