package gui.panel;

import gui.MainFrame;
import gui.dialog.PropertiesDialog;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;

import javax.swing.BorderFactory;
import javax.swing.DefaultCellEditor;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.ListSelectionModel;
import javax.swing.table.TableColumn;

import control.Control;
import control.Operation;

/**
 * Class implementing the process selection panel. Here the user can select 
 * which operations that should be part of the process.  
 * 
 * @author Kjell-Inge Skogstad and Trond ivind Eriksen
 */
public class ProcessPanel extends JPanel {

    private Control control;
    private JButton add, remove, edit;
    private ProcessModel model;
    private JTable table;
    
    /**
     * Constructor creating a new instance of this panel.
     * 
     * @param control The control instance
     */
    public ProcessPanel(Control control) {
        super(new BorderLayout());
        
        this.control = control;
        
        addComponents();
        addListeners();
        
        setBorder(BorderFactory.createTitledBorder("Process Operations"));
    }

    /**
     * Method adding all components.
     */
    private void addComponents() {
        JScrollPane scrollPane = new JScrollPane(
                JScrollPane.VERTICAL_SCROLLBAR_ALWAYS, 
                JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
        TableColumn column;
        JPanel buttonsPanel = new JPanel(new GridLayout(1, 3));
        JComboBox comboBox = new JComboBox();
        ArrayList tasks;
        
        /* The process table */
        table = new JTable(model = new ProcessModel());
        table.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
        for (int j = 0; j < model.getColumnCount(); j++) {
            column = table.getColumnModel().getColumn(j);
                    
            if (j == 0) {
                column.setPreferredWidth(10); 
            } else if (j == 1) {
                column.setPreferredWidth(150);
            } else {
                column.setPreferredWidth(10);
            }
        }
        
        /* Adding operations to the combobox */
        column = table.getColumnModel().getColumn(1);
        tasks = model.getPossibleOperations();
        for (int i = 0; i < tasks.size(); i++) {
            comboBox.addItem((String)tasks.get(i));
        }
        
        column.setCellEditor(new DefaultCellEditor(comboBox));
        
        scrollPane.setViewportView(table);
        scrollPane.setPreferredSize(new Dimension(200, 250));
        add(scrollPane, BorderLayout.NORTH);
        
        /* The buttons */
        buttonsPanel.add(add = new JButton("Add"));
        buttonsPanel.add(edit = new JButton("Edit"));
        buttonsPanel.add(remove = new JButton("Remove"));
        add(buttonsPanel, BorderLayout.CENTER);
    }
    
    /**
     * Method adding all listeners to the components.
     */
    private void addListeners() {
        add.addActionListener(new ActionListener() { 
            public void actionPerformed(ActionEvent e) {     
                int selected = table.getSelectedRow() + 1;
                
                if (model.getRowCount() == 0 && 
                        table.getSelectedRow() == 0) {
                    selected = -1;
                }
                model.addOperation(selected);
                
                /* Update table */
                table.getSelectionModel().setSelectionInterval(selected, 
                        selected);
                table.editingStopped(null);
                table.revalidate();
            }
        });
        edit.addActionListener(new ActionListener() { 
            public void actionPerformed(ActionEvent e) {  
                int selected = table.getSelectedRow();
                Operation op = model.getOperation(selected);
                
                if (op.getProperties() != null && op.getProperties().size() > 0) {
                    new PropertiesDialog(op);
                } else {
                    MainFrame.getInstance().printMessage("No properties available.");
                }
            }
        });
        remove.addActionListener(new ActionListener() { 
            public void actionPerformed(ActionEvent e) { 
                int selected = table.getSelectedRow();
      
                if (model.getRowCount() == 0) {
                    return;
                }
                model.removeOperation(selected); 
                
                /* Update table */
                table.getSelectionModel().setSelectionInterval(selected - 1, 
                            selected - 1);
                table.editingStopped(null);
                table.revalidate();
            }
        });
    }

    /**
     * Method used to retrieve the selected process.
     * 
     * @return The process
     */
    public ArrayList getProcess() {
        return model.getProcess();
    }
}
