package modelling.fiti;

import gui.MainFrame;
import gui.ReportView;
import gui.MainFrameDummy;
import gui.dialog.ReportDialog;

import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.GregorianCalendar;
import java.io.File;

import control.OperationProperty;
import control.Globals;

import modelling.ModelOperation;
import modelling.fiti.data.InterItem;
import modelling.fiti.data.Item;
import modelling.fiti.data.ItemSet;
import modelling.fiti.data.TermList;
import modelling.fiti.data.Transaction;
import modelling.fiti.helpers.StatusHelper;
import modelling.fiti.report.RuleModel;
import modelling.fiti.report.FitiReport;
import datamodel.*;

/**
 * Class implementing the FITI algorithm. This algorithm performs 
 * intertransaction association rules. 
 * 
 * Utilizes the FPgrowth algorithm to create large itemsets. 
 * 
 * @author Kjell-Inge Skogstad, Trond ivind Eriksen and Ole Kristian Fivelstad
 */
public class Fiti extends ModelOperation {

    //Debug
    static boolean DEBUG = false;

    //Default values of the operation properties
    private double minSupport = 0.1;
    private double maxSupport = 0.5;
    private double minConfidence = 0.5;
    private double maxConfidence = 1;
    private int maxspan = 3;
    private int maxSetSize = 3;
    private boolean pruneRulesByOrderOfTime = true;
    private int kFirst = -1;
    private boolean showStatus = true;


    private SimpleFiti simpleFiti;
    private TermList termList;


    /**
     * Method for performing the operation.
     */
    public void performOperation(DataSet dataSet) {
        ArrayList transactions;
        StatusHelper status = null;

        if(!Globals.showGui){
            showStatus=false;
        } else{
            MainFrame.getInstance().invalidOperationProgress();
        }
        //Test validity of dataSet
        if (!dataSet.termExtracted()) {
            if(Globals.showGui){
                MainFrame.getInstance().printMessage("Terms not available");
            } else {
                System.out.println("Terms not available");
            }
            return;
        }

        //Test if memory status should be shown
        if (showStatus) {
            status = new StatusHelper();
        }

        //Prints the settings to console
        System.out.println(printFitiSettings());
        

        //Import data
        termList = new TermList();
        transactions = createTransactions(dataSet, kFirst);
        System.out.println("Transactions created: " + transactions.size());
        System.out.println("Nof terms: " + termList.nofTerms());

        //Run Fiti
        simpleFiti = new SimpleFiti(transactions, termList,
                minSupport, maxSupport, minConfidence, maxConfidence, maxspan,
                maxSetSize, pruneRulesByOrderOfTime);


        //Stop the memory status thread if needed
        if (showStatus) {
            status.finish();
        }


    }

    /**
     * Method for printing the operation properties.
     * 
     * @return The printable properties
     */
    private String printFitiSettings() {
        String report = "";

        report += "FITI Settings: \n";
        report += "Minimum support: " + minSupport + "\n";
        report += "Maximum support: " + maxSupport + "\n";
        report += "Minimum confidence: " + minConfidence + "\n";
        report += "Maximum confidence: " + maxConfidence + "\n";
        report += "Maxspan: " + maxspan + "\n";
        report += "Max set size: " + maxSetSize + "\n";
        report += "Use only K first documents: " + kFirst + "\n";
        report += "Prune rules by order of time: " + pruneRulesByOrderOfTime + "\n\n";

        return report;
    }

    /**
     * Method for converting the dataset into transactions.
     * 
     * @param dataSet The dataset
     * @param kFirst K First
     * @return The list of transactions
     */
    private ArrayList createTransactions(DataSet dataSet, int kFirst) {
        Text text;
        ArrayList transactions = new ArrayList();
        String term;
        int currTime = ((Text)dataSet.getTexts().get(0)).getTid();
        Transaction t = new Transaction(currTime);

        //Test kFirst
        if (kFirst < 0 || kFirst > dataSet.getNofTexts()) {
            kFirst = dataSet.getNofTexts();
        }

        for (int i = 0; i < kFirst; i++) {
            text = (Text)dataSet.getTexts().get(i);


            //Test wether to create a new transaction or not.
            if (currTime != text.getTid()) {
                transactions.add(t);
                t = new Transaction(text.getTid());
                currTime = text.getTid();
            }
            for (int j = 0; j < text.getTerms().size(); j++) {
                term = ((Term)text.getTerms().get(j)).toString();
                termList.addTerm(term);
                Item item = new Item(termList.indexOf(term), Item.TERMLIST_INDEX);
                t.addItem(item);
            }
        }
        transactions.add(t);
        return transactions;
    }

    /**
     * Method for printing the results of the operation. Creates a ReportDialog.
     */
    public void printResults() {
        String report = "";
        DecimalFormat df = new DecimalFormat("0.00");
        ArrayList rules = simpleFiti.getRules();
        ArrayList largeInterSets = simpleFiti.getLargeSets();
        ItemSet set;

        if (largeInterSets != null) {
            //Settings
            report += printFitiSettings();

            //Large itemsets
            report += "Large itemsets: \n";
            for (int i = 0; i < largeInterSets.size(); i++) {
                set = (ItemSet)largeInterSets.get(i);

                report += "Set " + (i + 1) + ":   {";
                for (int j = 0; j < set.size(); j++) {
                    report += simpleFiti.getPrintableItem((InterItem)set.getItem(j)) + " ";
                }

                report += "},   support: " +
                    df.format(((ItemSet)largeInterSets.get(i)).getSupport()) + "\n";
            }
            report += "\n";
            if(Globals.showGui){
                new ReportDialog("FITI Results", new FitiReport(report, new RuleModel(rules)));
            }
            else{
                ReportView comp = new FitiReport(report, new RuleModel(rules));
                File resultDirectory = new File("results");
                if(!resultDirectory.exists()){
                    resultDirectory.mkdir();
                }
                String fileName="";
                Calendar now = new GregorianCalendar();

                int day = now.get(Calendar.DAY_OF_MONTH);
                String dayString = "";
                if(day<10){
                    dayString="0";
                }
                dayString+=(day+"");

                int month = now.get(Calendar.MONTH)+1;
                String monthString = "";
                if(month<10){
                    monthString="0";
                }
                monthString+=(month+"");

                int second = now.get(Calendar.SECOND);
                String secondString = "";
                if(second<10){
                    secondString="0";
                }
                secondString+=(second+"");

                int minute = now.get(Calendar.MINUTE);
                String minuteString = "";
                if(minute<10){
                    minuteString="0";
                }
                minuteString+=(minute+"");

                int hour = now.get(Calendar.HOUR_OF_DAY);
                String hourString = "";
                if(hour<10){
                    hourString="0";
                }
                hourString+=(hour+"");

                String time = dayString+monthString+"_"+hourString+minuteString+secondString+".txt";
                fileName=resultDirectory+File.separator+"Results_"+time;
                //storing rules
                MainFrameDummy.getInstance().getControl().saveReport(fileName,comp.getPrintableCopy());
            }
        }
    }

    /**
     * Method for getting the properties.
     */
    public ArrayList getProperties() {
        ArrayList properties = new ArrayList();

        properties.add(new OperationProperty("Minumum support", minSupport + ""));
        properties.add(new OperationProperty("Maximum support", maxSupport + ""));
        properties.add(new OperationProperty("Minumum confidence", minConfidence + ""));
        properties.add(new OperationProperty("Maximum confidence", maxConfidence + ""));
        properties.add(new OperationProperty("Maxspan", maxspan + ""));
        properties.add(new OperationProperty("Maximum set size", maxSetSize + ""));
        properties.add(new OperationProperty("Use K first documents", kFirst + ""));
        properties.add(new OperationProperty("Prune rules by order of time",
                new Boolean(pruneRulesByOrderOfTime).toString()));
        properties.add(new OperationProperty("Show Status",
                new Boolean(showStatus).toString()));

        return properties;
    }

    /**
     * Method for setting the properties.
     */
    public void setProperties(ArrayList properties) {
        OperationProperty p;

        for (int i = 0; i < properties.size(); i++) {
            p = (OperationProperty)properties.get(i);

            try {
                if (p.getName().equals("Minumum support")) {
                    minSupport = Double.parseDouble(p.getValue());
                } else if (p.getName().equals("Maximum support")) {
                    maxSupport = Double.parseDouble(p.getValue());
                } else if (p.getName().equals("Minumum confidence")) {
                    minConfidence = Double.parseDouble(p.getValue());
                } else if (p.getName().equals("Maximum confidence")) {
                    maxConfidence = Double.parseDouble(p.getValue());
                } else if (p.getName().equals("Maxspan")) {
                    maxspan = Integer.parseInt(p.getValue());
                } else if (p.getName().equals("Use K first documents")) {
                    kFirst = Integer.parseInt(p.getValue());
                } else if (p.getName().equals("Maximum set size")) {
                    maxSetSize = Integer.parseInt(p.getValue());
                } else if (p.getName().equals("Prune rules by order of time")) {
                    pruneRulesByOrderOfTime = Boolean.parseBoolean(p.getValue());
                } else if (p.getName().equals("Show Status")) {
                    showStatus = Boolean.parseBoolean(p.getValue());
                }
            } catch (Exception e) {
                System.out.println("Error parsing property " + p.getName());
            }
        }
    }
}
