package modelling.fiti;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;
import java.io.File;

import laur.dm.ar.AssociationRule;
import laur.dm.ar.Itemset;
import modelling.fiti.data.Filt;
import modelling.fiti.data.FitTable;
import modelling.fiti.data.InterItem;
import modelling.fiti.data.Item;
import modelling.fiti.data.ItemList;
import modelling.fiti.data.ItemSet;
import modelling.fiti.data.TermList;
import modelling.fiti.data.Transaction;
import modelling.fiti.helpers.DataSetHelper;

import external.modified.Apriori;
import external.modified.AprioriRules;
import control.Globals;
import shef.nlp.wordnet.similarity.SimilarityMeasure;
import net.didion.jwnl.data.POS;
import net.didion.jwnl.data.IndexWord;
import net.didion.jwnl.data.Synset;
import net.didion.jwnl.dictionary.Dictionary;
import net.didion.jwnl.JWNLException;

/**
 * The class implementing the FITI algorithm. This implementation is based on 
 * the article: Efficient Mining of Intertransaction Association Rules, by A K H Tung, 
 * H lu, J Han and L Feng (2003).
 * 
 * @author Kjell-Inge Skogstad, Trond ivind Eriksen and Ole Kristian Fivelstad
 */
public class SimpleFiti {

    private double minSupport;
    private double maxSupport;
    private double maxConfidence;
    private int maxspan;
    private int maxSetSize;
    private boolean pruneRulesByOrderOfTime=true;

    private SimilarityMeasure simMeasure;

    private ItemList itemList;
    private ArrayList currentInterLargeSets;
    private TermList termList;
    private Filt filt;
    private ArrayList fit;
    private ArrayList candidates;
    private ArrayList rules;

    /**
     * Constructor initializing and defining the algorithm.
     * 
     * @param transactions The transactions
     * @param termList The list of terms
     * @param minSupport Minimum support
     * @param maxSupport Maximum support
     * @param minConfidence Minimum confidence
     * @param maxConfidence Maximum confidence
     * @param maxspan Maxspan
     * @param maxSetSize Maximum set size
     * @param pruneRulesByOrderOfTime Pruning rules by order of time
     */
    public SimpleFiti(ArrayList transactions, TermList termList,
                      double minSupport, double maxSupport, double minConfidence,
                      double maxConfidence, int maxspan,
                      int maxSetSize, boolean pruneRulesByOrderOfTime) {
        int maxK;
        ArrayList currentLargeSets;

        //store variables
        this.termList = termList;
        this.minSupport = minSupport;
        this.maxSupport = maxSupport;
        this.maxConfidence = maxConfidence;
        this.maxspan = maxspan;
        this.maxSetSize = maxSetSize;
        this.pruneRulesByOrderOfTime = pruneRulesByOrderOfTime;

        itemList = new ItemList();
        currentInterLargeSets = new ArrayList();

        //find large intra sets 
        currentLargeSets = findLargeIntraSets(transactions);
        System.out.println("Nof large items:" + currentLargeSets.size());
        //Create FILT structure
        maxK = createFilt(currentLargeSets);
        System.out.println("Nof itemsets in FILT: " + filt.size());
        System.out.println(filt.toString());


        //Fill FIT tables
        fillFit(maxK, transactions);
        System.out.println("Nof FIT tables: " + fit.size());
        for (int i = 0; i < fit.size(); i++) {
            debug("FIT table " + (i + 1) + ": " + ((FitTable)fit.get(i)).toString());
        }
        //Run levelwise APRIORI to generate large inter sets
        interApriori();
        System.out.println("Nof large inter sets: " + currentInterLargeSets.size());
        for (int i = 0; i < currentInterLargeSets.size(); i++) {
            debug(((ItemSet)currentInterLargeSets.get(i)).toString());
        }

        //Create rules
        debug("Find rules");
        ArrayList rules = new AprioriRules().findAssociations(getItemsets(),
                minSupport, minConfidence);


        //
        Map<String,String> params = new HashMap<String,String>();
        params.put("simType","shef.nlp.wordnet.similarity.JCn");
        params.put("infocontent","file:jwnl"+File.separator+"ic-bnc-resnik-add1.dat");
        //params.put("mapping","file:test/domain_independent.txt");
        try{
            simMeasure = SimilarityMeasure.newInstance(params);
        } catch (Exception e) {
            System.out.println("Similarity Measure not initialized");
            simMeasure=null;
            //e.printStackTrace();
        }


        // Insert rules into own model
        this.rules = new ArrayList();
        for (int i = 0; i < rules.size(); i++) {
            storeRule((AssociationRule)rules.get(i));            
        }
        System.out.println("\nFiti finished " + this.rules.size() + " rules");
    }

    /**
     * Method for validating duplicity of the itemset. Checks if the 
     * candidate itemset is already present.
     * 
     * @param set The itemset
     * @param candidates The candidates
     * @return The new itemset (null if already present)
     */
    private ItemSet validateDuplicity(ItemSet set, ArrayList candidates) {
        ItemSet temp;

        for (int i = 0; i < candidates.size(); i++) {
            temp = (ItemSet)candidates.get(i);

            if (temp.equals(set)) {
                return null;
            }
        }
        return set;
    }

    /**
     * Method for validating the filt index of the itemset. Checks and updates 
     * the filt index of the itemset. 
     * 
     * @param set The itemset
     * @return The new itemset
     */
    private ItemSet validateFiltIndex(ItemSet set) {
        ItemSet validated = new ItemSet();
        ItemSet clone;
        ItemSet newClone;
        ItemSet temp;
        InterItem item=null;
        ArrayList list;

        //Test for each point in time
        for (int i = 0; i < maxspan; i++) {;
            clone = set.clone();
            newClone = new ItemSet();

            //Extract items with given timestamp
            for (int j = 0; j < clone.size(); j++) {
                item = (InterItem)clone.getItem(j);

                if (item.getTimestamp() == i) {
                    newClone.addItem(item);
                }
            }

            //Create a new itemset and check filt index
            if (newClone.size() > 0) {
                list = new ArrayList();
                clone = new ItemSet();

                //Rebuild a new set 
                for (int j = 0; j < newClone.size(); j++) {
                    temp = new ItemSet();
                    item = (InterItem)newClone.getItem(j);
                    temp.addItem(new Item(item.getIndex(), item.getIndexType()));
                    temp = generateFiltVersion(temp);
                    list.addAll(filt.getAllObjectsIn(filt.findIndexOfSet(temp)));
                    //debug("Add  til list " + filt.getAllObjectsIn(temp).toString() + " where temp is " + temp.toString());
                }
                //debug("Liste " + list.toString());

                for (int j = 0; j < list.size(); j++) {
                    clone.addItem(new Item(((Integer)list.get(j)).intValue(),Item.FILT_INDEX));
                }

                //Find filt index
                clone = generateFiltVersion(clone);
                Item it = itemList.getItem(Integer.valueOf((clone.getItems().get(0)).toString()));
                validated.addItem(new InterItem(filt.findIndexOfSet(clone),
                        Item.FILT_INDEX , i));
            }
        }
        return validated;
    }

    /**
     * Method for getting a printable version of an item.
     * 
     * @param item The item
     * @return The string representation
     */
    public String getPrintableItem(InterItem item) {
        String printable = "";
        Item it;
        ArrayList objectSets;
        ItemSet itemSet = new ItemSet();
        itemSet.addItem(new Item(item.getIndex(), item.getIndexType()));
        itemSet = generateFiltVersion(itemSet);

        objectSets = filt.getAllObjectsIn(filt.findIndexOfSet(itemSet));

        printable += "(";
        for (int j = 0; j < objectSets.size(); j++) {
            it = filt.getItemSet(((Integer)objectSets.get(j)).intValue()).getItem(0);
            printable += "'" + termList.getTerm(it.getIndex()) + "' ";
        }
        printable += ", " + item.getTimestamp() + ") ";

        return printable;
    }

    /**
     * Method for storing a given rule in the local data structure.
     * 
     * @param rule The rule
     */
    private void storeRule(AssociationRule rule) {
        String left = "";
        String right = "";
        ArrayList leftItems =new ArrayList();
        ArrayList rightItems =new ArrayList();
        boolean valid=false;
        int time = -1;
        //Test confidence
        if (rule.getConfidence() <= maxConfidence ) {
            //Build left side
            for (int i = 0; i < rule.antecedentSize(); i++) {

                if(((InterItem)itemList.getItem(rule.getAntecedentItem(i))).getTimestamp()==0){
                    valid = true;
                }
                // Update max time
                if (pruneRulesByOrderOfTime &&
                        time < ((InterItem)itemList.getItem(rule.getAntecedentItem(i))).getTimestamp()) {
                    time = ((InterItem)itemList.getItem(rule.getAntecedentItem(i))).getTimestamp();
                }
                InterItem item = (InterItem)itemList.getItem(rule.getAntecedentItem(i));
                left += getPrintableItem((InterItem)itemList.getItem(rule.getAntecedentItem(i)));

                String temp = removeChars(getPrintableItem(item));
                temp=temp.substring(0, temp.indexOf(","));
                String words[] = temp.split(" ");
                for(int j=0; j<words.length; j++){
                    if(words[j].contains("#")) {
                            leftItems.add(words[j].trim());
                        } /*else if(words[j].contains("COLL")&&words[j].contains("#")){
                            String word = words[j];
                            rightItems.add(word.trim());
                        } */
                }
            }
            //Build right side
            if(valid){
                for (int i = 0; i < rule.consequentSize(); i++) {
                    // Test on time
                    if (pruneRulesByOrderOfTime &&
                            time > ((InterItem)itemList.getItem(rule.getConsequentItem(i))).getTimestamp()) {
                        return;
                    }
                    right += getPrintableItem((InterItem) itemList.getItem(rule.getConsequentItem(i)));
                    InterItem item = (InterItem) itemList.getItem(rule.getConsequentItem(i));

                    String temp = removeChars(getPrintableItem(item));
                    temp=temp.substring(0, temp.indexOf(","));
                    String words[] = temp.split(" ");
                    for(int j=0; j<words.length; j++){
                        if(words[j].contains("#")) {
                            rightItems.add(words[j].trim());
                        }/* else if(words[j].contains("COLL")&&words[j].contains("#")){
                            String word = words[j];
                            rightItems.add(word.trim());
                        }*/
                    }

                }

                //Calculate semantic similarity
                double semanticSimilarity;
                if(leftItems.size()==0||rightItems.size()==0){
                    semanticSimilarity=0;
                } else{
                    semanticSimilarity=calculateSimilarity(leftItems, rightItems);
                }

                //Add rule
                rules.add(new Rule("{" + left + "} -> {" + right + "}",
                            rule.getSupport(), rule.getConfidence(), semanticSimilarity));
            }

        }
    }


    /**
     * Method for calculating the average semantic similarity between
     * the items on the left and right side of the rule.
     *
     * @param leftSide The items on the left side
     * @param rightSide The items on the right side
     * @return The similarity
     */
    private double calculateSimilarity(ArrayList leftSide, ArrayList rightSide){
        double result=0;
        double total=0;
        int comparisons=0;
        if(simMeasure==null) return 0;
        for(int i=0; i<leftSide.size();i++){
            String leftWord = leftSide.get(i).toString();
            if(leftWord.contains("/")){
                String temp = leftWord.substring(0, leftWord.indexOf("/"));
                leftWord = temp+leftWord.substring(leftWord.indexOf("#"), leftWord.length());
            }
            int sense=findSense(leftWord);
            try {
                IndexWord left = Dictionary.getInstance().getMorphologicalProcessor()
                        .lookupBaseForm(POS.NOUN, leftWord.toLowerCase());
                Synset leftSynset = left.getSense(sense);
                for(int j=0; j<rightSide.size(); j++){
                    String rightWord = rightSide.get(j).toString();
                    if(rightWord.contains("/")){
                        String temp = rightWord.substring(0, rightWord.indexOf("/"));
                        rightWord = temp+rightWord.substring(rightWord.indexOf("#"), rightWord.length());
                    }
                    int sense2 = findSense(rightWord);
                    IndexWord right = Dictionary.getInstance().getMorphologicalProcessor()
                            .lookupBaseForm(POS.NOUN, rightWord.toLowerCase());
                    Synset rightSynset = right.getSense(sense2);
                    double sim=simMeasure.getSimilarity(leftSynset, rightSynset);
                    total+=sim;
                    comparisons++;
                }
            } catch (JWNLException e) {
                e.printStackTrace();
            }

        }
        result = total/comparisons;      

        return result;
    }

    /**
     * Method for getting the sense of a word.
     * E.g. ball#1 returns 1
     * @param word The word
     * @return The sense
     */
    private int findSense(String word){
        int sense=1;
        String senseString = word.substring(word.indexOf("#")+1, word.length());
        sense = Integer.parseInt(senseString);

        return sense;
    }

    /**
     * Method for removing specific characters from a string.
     *
     * @param item The string
     * @return The new string
     */
    private String removeChars(String item){
        item=item.replaceAll("\\(","");
        item=item.replaceAll("\\)","");
        item=item.replaceAll("'","");
        return item;
    }

    /**
     * Method for getting all large itemsets, converted to Itemset
     * 
     * @return List of large itemsets
     */
    private ArrayList getItemsets() {
        ArrayList list = new ArrayList();

        for (int i = 0; i < currentInterLargeSets.size(); i++) {
            list.add(DataSetHelper.convertFromItemSet((ItemSet)currentInterLargeSets.get(i)));
        }
        return list;
    }

    /**
     * Method for creating and filling the Filt structure.
     * 
     * @param currentLargeSets Large itemsets
     * @return Size of largest itemset
     */
    private int createFilt(ArrayList currentLargeSets) {
        int candMax = 0;
        int max = 0;
        ItemSet set;

        filt = new Filt();

        for (int i = 0; i < currentLargeSets.size(); i++) {
            set = (ItemSet)currentLargeSets.get(i);

            if (set.size() <= maxSetSize) {
                filt.addItemSet(set);
                candMax = set.size();

                //Find largest large itemset
                if (candMax > max) {
                    max = candMax;
                }
            }
        }

        debug("Updating pointers");
        filt.updatePointers();
        return max;
    }

    /**
     * Method implementing part 3 of the algorithm. APRIORI is run levelwise to 
     * generate large inter itemsets.
     */
    private void interApriori() {
        ArrayList firstItems;
        ArrayList pointers;
        ArrayList megaTransactions;
        int k = 1;

        //Analyse time
        pointers = analyseTime();
        debug("Nof windows: " + pointers.size());
        for (Object integer : pointers) {
            debug("Start at: " + (Integer)integer);
        }

        //Create mega transactions
        megaTransactions = createMegaTransactions(pointers);
        System.out.println("Nof mega transactions:" + megaTransactions.size());
        for (Object t : megaTransactions) {
            debug(((Transaction)t).toString());
        }

        //Create all time combinations of candidates
        candidates = new ArrayList();
        firstItems = filt.getFirstItemSets();
        debug("First itemsets:");
        debug(firstItems.toString());

        for (int i = 0; i < firstItems.size(); i++) {
            candidates.addAll(generateCombinations(((ItemSet)firstItems.get(i)).getItem(0).getIndex()));
        }
        debug("Nof 1-candidates: " + candidates.size());
        debug(candidates.toString());

        //Update count and generate new candidates
        while (candidates.size() > 0) {
            //Update count
            System.out.println("Updating count on level " + k + ", " +
                    candidates.size() + " candidates");
            updateCount(megaTransactions);

            //Store large
            storeLarge(candidates, pointers.size());

            k++;
            if (k > maxSetSize) {
                break;
            }

            //Generate next candidates
            System.out.println("Generate next candidates, k = " + k);
            generateNextCandidates();
            debug("Nof " + k + "-candidates: " + candidates.size());
            debug(candidates.toString());
        }
    }

    /**
     * Method for creating the mega transactions.
     * 
     * @param pointers List of pointers to start transactions 
     * of each mega transaction
     * @return List of mega transactions
     */
    private ArrayList createMegaTransactions(ArrayList pointers) {
        ArrayList megaTransactions = new ArrayList();
        Transaction t;
        FitTable fitTable = (FitTable)fit.get(0);;
        int start = 0;
        int end = 0;
        int time;
        int currTime = 0;
        int counter;
        int distance;
        ArrayList filtIndexes;

        //For each mega transaction...
        for (int i = 0; i < pointers.size(); i++) {
            start = (Integer)pointers.get(i);
            counter = 0;
            distance = 0;
            currTime = fitTable.getTimestamp(start);

            //Find correct end pointer 
            while (true) {
                counter++;

                if (counter >= fitTable.size()) {
                    end = fitTable.size();
                    break;
                }

                if (currTime != fitTable.getTimestamp(counter)) {
                    distance += fitTable.getTimestamp(counter) - currTime;
                    currTime = fitTable.getTimestamp(counter);
                }

                if (distance >= maxspan) {
                    end = counter;
                    break;
                }
            }

            t = new Transaction();

            //fill transactions with itemsets of inter items
            for (int j = 0; j < fit.size(); j++) {
                fitTable = (FitTable)fit.get(j);
                currTime = 0;
                time = fitTable.getTimestamp(start);

                for (int k = start; k < end; k++) {
                    filtIndexes = fitTable.getSets(k);

                    if (time != fitTable.getTimestamp(k)) {
                        time = fitTable.getTimestamp(k);
                        currTime++;
                    }

                    for (int l = 0; l < filtIndexes.size(); l++) {
                        t.addItem(new InterItem((Integer)filtIndexes.get(l),
                                Item.FILT_INDEX, currTime));
                    }
                }
            }
            megaTransactions.add(t);
        }
        return megaTransactions;
    }

    /**
     * Method for updating the support count of candidate itemsets.
     * 
     * @param megaTransactions The mega transactions
     */
    private void updateCount(ArrayList megaTransactions) {
        Transaction t;
        ItemSet itemSet;

        for (int i = 0; i < megaTransactions.size(); i++) {
            t = (Transaction)megaTransactions.get(i);

            //Test if candidates are present in mega transaction
            for (int j = 0; j < candidates.size(); j++) {
                itemSet = (ItemSet)candidates.get(j);
                //debug("transaction: " + t.toString() + ", itemset: " + itemSet.toString());
                if (t.contains(itemSet)) {
                    itemSet.addOne();
                }
            }
        }
    }

    /**
     * Method for storing large inter itemsets.
     * 
     * @param candidatesWithCount List of candidates
     * @param nofMega Number of mega transactions
     */
    private void storeLarge(ArrayList candidatesWithCount, int nofMega) {
        ItemSet set;
        ArrayList temp = new ArrayList();

        for (int i = 0; i < candidatesWithCount.size(); i++) {
            set = (ItemSet)candidatesWithCount.get(i);
            set.setSupport((double)set.getCount() / (double)nofMega);

            //Test on minimum support
            if (set.getSupport() >= minSupport) {
                currentInterLargeSets.add(set);
                temp.add(set);
                debug("Storing: " + set + ", count = " + set.getCount());
            }
        }
        candidates = temp;
    }

    /**
     * Method for analysing time and finding starting pointers of all mega
     * transactions.
     * 
     * @return List of starting pointers
     */
    private ArrayList analyseTime() {
        ArrayList pointers = new ArrayList();
        FitTable table = (FitTable)fit.get(0);
        int endTime = table.getTimestamp(table.size() - 1);
        int currTime = table.getTimestamp(0);
        int counter = 0;
        int windowTime = 0;
        int first = 0;

        pointers.add(new Integer(0));
        while (currTime < endTime) {
            if (currTime != table.getTimestamp(counter)) {
                if (first == 0) {
                    first = counter;
                }

                windowTime += table.getTimestamp(counter) - currTime;
                currTime = table.getTimestamp(counter);
            }

            if (windowTime >= maxspan) {
                pointers.add(new Integer(first));
                windowTime = 0;
                counter = first;
                currTime = table.getTimestamp(counter);
                first = 0;
            }
            counter++;
        }

        return pointers;
    }

    /**
     * Method for generating the next candidates based on current
     * large inter itemsets.
     */
    private void generateNextCandidates() {
        ArrayList newCandidates = new ArrayList();
        ArrayList children;
        ItemSet set;
        ArrayList comb;
        ItemSet combSet;

        for (int i = 0; i < candidates.size(); i++) {
            set = generateFiltVersion((ItemSet)candidates.get(i));

            //Find children
            children = filt.getChildrenOf(filt.findIndexOfSet(set));
            debug("Children of " + set.toString() + " are " + children.toString());

            //Create candidates
            for (int j = 0; j < children.size(); j++) {
                comb = generateCombinations(filt.getItemSet(((Integer)children.get(j)).intValue()),
                        (ItemSet)candidates.get(i));
                debug("Combinations " + comb.toString());

                for (int k = 0; k < comb.size(); k++) {
                    combSet = (ItemSet)comb.get(k);
                    debug("Unvalidated: " + combSet);

                    combSet = validateFiltIndex(combSet);
                    //debug("vf: " + combSet);

                    if (combSet != null) {
                        combSet = validateDuplicity(combSet, newCandidates);
                        //debug("vd: " + combSet);

                        if (combSet != null) {
                            newCandidates.add(combSet);
                        } else {
                            debug("Duplicate exist");
                        }
                    } else {
                        debug("Not exist in filt ");
                    }
                }
            }
        }

        //debug("New candidates ");
        //debug(newCandidates.toString());
        candidates = newCandidates;
    }

    /**
     * Method for generating all combinations (time) of a itemset, given its
     * parent inter itemset. (Only one item has no specific timestamp. Therefore
     * all combinations of this item is created with corresponding other items.)
     * 
     * @param intraChild The itemset
     * @param interParent The parent inter itemset
     * @return List of combinations
     */
    private ArrayList generateCombinations(ItemSet intraChild,
                                           ItemSet interParent) {
        ItemSet interParentFilt = generateFiltVersion(interParent);
        ArrayList combinations = new ArrayList();
        ItemSet intraClone = intraChild.clone();
        ItemSet interClone;
        Item item;
        Item itemFilt;
        ArrayList combTime;

        //Identifying item without timestamp
        for (int i = 0; i < interParentFilt.size(); i++) {
            for (int j = 0; j < intraClone.size(); j++) {
                itemFilt = interParentFilt.getItem(i);
                item = intraClone.getItem(j);
                if (itemFilt.equals(item)) {
                    intraClone.removeItem(j);
                    break;
                }
            }
        }

        //Generate combination of itemset
        combTime = generateCombinations(filt.findIndexOfSet(intraClone));

        //Rebuild inter itemset
        for (int i = 0; i < combTime.size(); i++) {
            interClone = interParent.clone();
            interClone.addItem(((ItemSet)combTime.get(i)).getItem(0));
            combinations.add(interClone);
        }
        return combinations;
    }

    /**
     * Method for generating all combinations (time) of a given itemset.
     * 
     * @param filtIndex The index of the itemset
     * @return List of combinations
     */
    private ArrayList generateCombinations(int filtIndex) {
        ArrayList combinations = new ArrayList();
        ItemSet temp;

        //Add itemset for each timestamp
        for (int i = 0; i < maxspan; i++) {
            temp = new ItemSet();
            ItemSet iSet = filt.getItemSet(filtIndex);
            ArrayList items = iSet.getItems();
            Item it = itemList.getItem(Integer.valueOf((items.get(0)).toString()));
            temp.addItem(new InterItem(filtIndex, Item.FILT_INDEX, i));
            combinations.add(temp);
        }

        return combinations;
    }

    /**
     * Method used to create and fill the FIT structure. Algorithm given 
     * in article
     * 
     * @param maxK Maximum number of FIT tables.
     * @param transactions The transactions
     */
    private void fillFit(int maxK, ArrayList transactions) {
        fit = new ArrayList();

        for (int i = 0; i < maxK; i++) {
            fit.add(new FitTable());
        }
        transform(transactions);
    }

    /**
     * Method used to parse the transactions and fill the FIT structure with
     * Filt indexes. Algorithm given in article.
     * 
     * @param transactions The transactions
     */
    private void transform(ArrayList transactions) {
        Transaction t;

        for (int i = 0; i < transactions.size(); i++) {
            t = (Transaction)transactions.get(i);

            for (int j = 0; j < fit.size(); j++) {
                ((FitTable)fit.get(j)).addEntry(t.getTimestamp());
            }
            subset(t, 0, 1, 0);
        }
    }

    /**
     * Method used in filling the FIT tables. See transform(). 
     * Algorithm given in article.
     * 
     * @param t A transaction
     * @param index The index
     * @param k The level (size of itemset)
     * @param nodeId The filt index
     */
    private void subset(Transaction t, int index, int k, int nodeId) {
        ArrayList items = t.getItems();
        ItemSet set;
        int id;
        int nowId;
        ItemSet tempSet;
        Item tempItem;

        if (k == 1) {
            for (int i = index; i < items.size(); i++) {
                set = new ItemSet();
                tempItem = (Item)items.get(i);
                set.addItem(new Item(tempItem.getIndex(), tempItem.getIndexType()));
                set = generateFiltVersion(set);

                id = filt.findIndexOfSet(set);

                if (id != -1) {
                    nowId = id;
                    ((FitTable)fit.get(0)).addSet(t.getTimestamp(), nowId);

                    debug(k + ": Legger til: " + nowId + " med tid " + t.getTimestamp());
                    for (int j = i + 1; j < items.size(); j++) {
                        tempSet = set.clone();
                        tempItem = (Item)items.get(j);
                        tempSet.addItem(tempItem);
                        tempSet = generateFiltVersion(tempSet);

                        id = filt.findIndexOfSet(tempSet);

                        if (id != -1) {
                            subset(t, j + 1, k + 1, id);
                        }
                    }
                }
            }
        } else if (k <= fit.size()) {
            ((FitTable)fit.get(k - 1)).addSet(t.getTimestamp(), nodeId);

            debug(k + ": Legger til: " + nodeId + " med tid " + t.getTimestamp());
            for (int j = index; j < items.size(); j++) {
                tempItem = (Item)items.get(j);
                set = ((ItemSet)filt.getItemSet(nodeId)).clone();
                set.addItem(tempItem);
                set = generateFiltVersion(set);

                id = filt.findIndexOfSet(set);

                if (id != -1) {
                    subset(t, j + 1, k + 1, id);
                }
            }
        }
    }

    /**
     * Method used to find all large intra itemsets. Runs external algorithm 
     * (APRIORI with hashing) to generate itemsets.
     * 
     * @param transactions The transactions
     * @return List of large itemsets
     */
    private ArrayList findLargeIntraSets(ArrayList transactions) {
        ArrayList external =  new ArrayList();
        ArrayList currentLargeSets = new ArrayList();
        ItemSet set;
        ArrayList items;
        ArrayList largeSets;

        //Copy transactions to external data structure
        for (int i = 0; i < transactions.size(); i++) {
            items = ((Transaction)transactions.get(i)).getItems();
            set = new ItemSet();

            for (int j = 0; j < items.size(); j++) {
                set.addItem((Item)items.get(j));
            }
            external.add(DataSetHelper.convertFromItemSet(set));
        }

        transactions = null;
        System.gc();

        debug("Find large intra sets");
        //Run algorithm
        debug("Runinng extern algorithm apriori: minSupport = " + minSupport);
        largeSets = new Apriori().findFrequentItemsets(external,
                termList.nofTerms(), minSupport, maxSupport, maxSetSize);

        //Copy large sets into own data structure
        for (int i = 0; i < largeSets.size(); i++) {
            currentLargeSets.add(DataSetHelper.covertFromItemset((Itemset)largeSets.get(i)));
        }
        return currentLargeSets;
    }

    /**
     * Debug method used to print debug messages. Messages are only printed if
     * the static variable DEBUG is set to true.
     * 
     * @param string The message
     */
    private void debug(String string) {
        if (Fiti.DEBUG) {
            System.out.println(string);
        }
    }

    /**
     * Method used to get all large inter itemsets.
     * 
     * @return List of large inter itemsets.
     */
    public ArrayList getLargeSets() {
        return currentInterLargeSets;
    }

    /**
     * Method used to get all rules generated.
     * 
     * @return List of rules
     */
    public ArrayList getRules() {
        return rules;
    }

    /**
     * Method used to retrieve a filt version (intra version) of a itemset. 
     * Used in finding filt indexes.
     * 
     * @param itemSet The itemset
     * @return The filt version of the itemset
     */
    public ItemSet generateFiltVersion(ItemSet itemSet) {
        ItemSet newItemSet = new ItemSet();
        ArrayList subSet;
        Item old;
        int index;

        for (int i = 0; i < itemSet.size(); i++) {
            old = itemSet.getItem(i);

            if (old.getIndexType() == Item.FILT_INDEX) {
                subSet = filt.getAllObjectsIn(old.getIndex());

                for (int j = 0; j < subSet.size(); j++) {
                    index = ((Integer)subSet.get(j)).intValue();
                    newItemSet.addItem(filt.getItemSet(index).getItem(0));
                }
            } else {
                newItemSet.addItem(old);
            }
        }
        return newItemSet;
    }
}
