 package modelling.fiti.data;

import java.util.ArrayList;

/**
 * Class implementing the Filt structure described in article; Efficient Mining 
 * of Intertransaction Association Rules, by A K H Tung, H lu, J Han and L Feng 
 * (2003).
 * 
 * @author Kjell-Inge Skogstad
 */
public class Filt {

    private ArrayList<Node> nodes;

    /**
     * Constructor creatin a new FILT instance.
     */
    public Filt() {
        nodes = new ArrayList<Node>();
    }

    /**
     * Method for adding an itemset.
     * 
     * @param set The itemset
     */
    public void addItemSet(ItemSet set) {
        nodes.add(new Node(set));
    }  
    /**
     * Method for generating pointers between parents -> children, 
     * and parents -> elements, where elements are all single items in the 
     * itemset.
     */
    public void updatePointers() {
        Node parent;
        Node child;
        
        for (int i = 0; i < nodes.size(); i++) {
            parent = (Node)nodes.get(i);
            
            if (parent.itemset.size() == 1) {
                parent.elements.add(new Integer(i + 1));
            }
            
            for (int j = i + 1; j < nodes.size(); j++) {
                child = (Node)nodes.get(j);
                
                //update children list
                if (child.containSet(parent.itemset) && 
                        child.itemset.size() == parent.itemset.size() + 1) {
                    parent.children.add(new Integer(j + 1));
                }
                
                //update elements list
                if (child.containSet(parent.itemset) && parent.itemset.size() == 1) {
                    child.elements.add(new Integer(i + 1));
                } 
            }
        }
    }
    
    /**
     * Mehtod for getting an itemset.
     * 
     * @param index The index
     * @return The itemset
     */
    public ItemSet getItemSet(int index) {
        return ((Node)nodes.get(index - 1)).itemset;
    }
    
    /**
     * Method for gettin all itemsets of size 1. That is, the first itemsets.
     * 
     * @return List of itemsets
     */
    public ArrayList getFirstItemSets() {
        ArrayList<ItemSet> first = new ArrayList<ItemSet>();
        Node node;
        ItemSet set;
        
        for (int i = 0; i < nodes.size(); i++) {
            node = (Node)nodes.get(i);
            
            //Test on itemset size
            if (node.itemset.size() == 1) {
                set = new ItemSet();
                set.addItem(new Item(i + 1, Item.FILT_INDEX));
                first.add(set);
            }
        }
        return first;
    }

    /**
     * Method for getting all children of given itemset. That is, all itemsets
     * of size + 1, and which contain the given itemset as well.
     * 
     * @param index The index of parent itemset
     * @return List of children itemsets
     */
    public ArrayList getChildrenOf(int index) {
        index--;
        
        if (index >= 0 && index < nodes.size()) {
            return ((Node)nodes.get(index)).children;
        }      

        return new ArrayList();
    }
    
    /**
     * Method for getting the size of, the number itemsets in, the filt structure
     * 
     * @return Size
     */
    public int size() {
        return nodes.size();
    }
    
    /**
     * Method for gettin all items, itemsets of size 1, in given itemset.
     * 
     * @param index The index
     * @return List of itemsets
     */
    public ArrayList getAllObjectsIn(int index) {
        index--;
         
        if (index >= 0 && index < nodes.size()) {
            return ((Node)nodes.get(index)).elements;
        }       
        return new ArrayList();
    }
    
    /**
     * Method for getting the index of a given itemset. Note that the method only
     * handles itemsets where the items have termlist index. (Filt version).
     */
    public int findIndexOfSet(ItemSet set) {   
        if (set.getItem(0).getIndexType() == Item.TERMLIST_INDEX) {
            for (int i = 0; i < size(); i++) {
                //Test equivalence
                if (((Node)nodes.get(i)).itemset.equals(set)) {
                    return i + 1;
                }
            }
        }
        return -1;
    }
    
    /**
     * Methodd for getting a printable representation of the filt structure.
     */
    public String toString() {
        String string = "";
        
        for (int i = 0; i < nodes.size(); i++) {
            string += (i + 1) + " " + ((Node)nodes.get(i)).toString() + "\n";
        }
        return string;
    }
    
    /**
     * Internal class, representing a filt node. A filt node contains the 
     * itemset, a list of children and a list of elements.
     * 
     * @author Kjell-Inge Skogstad and Trond ivind Eriksen
     */
    private class Node {
        
        private ItemSet itemset;
        private ArrayList<Integer> children;
        private ArrayList<Integer> elements;        

        /**
         * Constructor creating a new filt node.
         * 
         * @param itemset A itemset
         */
        public Node(ItemSet itemset) {
            this.itemset = itemset;
            children = new ArrayList<Integer>();
            elements = new ArrayList<Integer>();
        }
        
        /**
         * Method for testing if the node contains an itemset.
         * 
         * @param set The itemset
         * @return True if node contains itemsets
         */
        public boolean containSet(ItemSet set) {
            int counter = 0;
            Item item;
            
            for (int i = 0; i < set.size(); i++) {
                item = new Item(set.getItem(i).getIndex(), set.getItem(i).getIndexType());
                
                if (itemset.containsItem(item)) {
                    counter++;
                }
            }
            return counter == set.size();
        }
        
        /**
         * Method for getting a printable representation of the node.
         */
        public String toString() {
            return itemset.toString() + " " + itemset.getSupport();
        }
    }
}
