package modelling.fiti.data;

import java.util.ArrayList;

/**
 * Class implementing a FIT table according to article; Efficient Mining 
 * of Intertransaction Association Rules, by A K H Tung, H lu, J Han and L Feng 
 * (2003). Each entry in the table consists of a list of itemsets and a 
 * timestamp.
 * 
 * Only contains itemsets with items with filt index.
 * 
 * @author Kjell-Inge Skogstad and Trond ivind Eriksen
 */
public class FitTable {

    private ArrayList table;
    
    /**
     * Constructor creating a new FIT table.
     */
    public FitTable() {
        table = new ArrayList();
    }
    
    /**
     * Method returning the number of entries in the FIT table.
     * 
     * @return Number of entries
     */
    public int size() {
        return table.size();
    }
    
    /**
     * Method for adding an entry with a given timestamp.
     * 
     * @param timestamp The timestamp
     */
    public void addEntry(int timestamp) {
        table.add(new TableEntry(timestamp));
    }
    
    /**
     * Method for adding an itemset to a specified entry in the table.
     * 
     * @param timestamp The timestamp
     * @param setId The index of the itemset
     */
    public void addSet(int timestamp, int setId) {
        ((TableEntry)table.get(findEntry(timestamp))).addSet(new Integer(setId));
    }
    
    /**
     * Method for getting the itemsets of an entry.
     * 
     * @param index The index
     * @return List of itemsets
     */
    public ArrayList getSets(int index) {
        return ((TableEntry)table.get(index)).sets;
    }
    
    /**
     * Method for getting the timestamp of an entry
     * 
     * @param index The index
     * @return The timestamp
     */
    public int getTimestamp(int index) {
        return ((TableEntry)table.get(index)).timestamp;
    }
    
    /**
     * Method for finding the index of an entry given its timestamp.
     * 
     * @param timestamp The timestamp
     * @return The index
     */
    private int findEntry(int timestamp) {
        for (int i = 0; i < table.size(); i++) {
            if (timestamp == ((TableEntry)table.get(i)).timestamp) {
                return i;
            }
        }
        return -1;
    }
    
    /**
     * Method for getting a printable representation of the FIT table.
     */
    public String toString() {
        return table.toString();
    }
    
    /**
     * Class implementing an entry in the FIT table. En entry includes a 
     * timestamp and a list of itemsets.
     * 
     * @author Kjell-Inge Skogstad and Trond ivind Eriksen
     */
    private class TableEntry {
        
        private int timestamp; 
        private ArrayList sets;
        
        /**
         * Constructor creating a new table entry.
         * 
         * @param timestamp The timestamp
         */
        public TableEntry(int timestamp) {
            this.timestamp = timestamp;
            sets = new ArrayList();
        }
        
        /**
         * Method for adding an itemset.
         * 
         * @param set
         */
        public void addSet(Object set) {
            sets.add(set);
        }
        
        /**
         * Method for getting a printable representation of the table entry.
         */
        public String toString() {
            return "(" + sets.toString() + ", " + timestamp + ")";
        }
    }
}
